"""Tests for coralsnake.genegroup – rename helpers and reassignment logic."""

import re

import pytest

from coralsnake.genegroup import rename_snRNA, rename_snoRNA


# ---------------------------------------------------------------------------
# rename_snRNA
# ---------------------------------------------------------------------------
class TestRenameSnRNA:
    def test_RNU2(self):
        assert rename_snRNA("RNU2-1") == "U2"

    def test_RNU2_pseudo(self):
        assert rename_snRNA("RNU2-27P") == "U2"

    def test_RNU4ATAC(self):
        assert rename_snRNA("RNU4ATAC7") == "U4atac"

    def test_RNVU1(self):
        assert rename_snRNA("RNVU1-1") == "U1V"

    def test_drosophila(self):
        assert rename_snRNA("snRNA:U2:38ABb") == "U2"

    def test_passthrough(self):
        assert rename_snRNA("unrelated") == "unrelated"


# ---------------------------------------------------------------------------
# rename_snoRNA
# ---------------------------------------------------------------------------
class TestRenameSnoRNA:
    def test_ACA(self):
        assert rename_snoRNA("ACA64") == "SNORA64"

    def test_U8(self):
        assert rename_snoRNA("U8") == "SNORD118"

    def test_U3(self):
        r = rename_snoRNA("U3")
        assert r.startswith("SNORD")

    def test_Z6(self):
        assert rename_snoRNA("Z6") == "Z6"


# ---------------------------------------------------------------------------
# Reassignment logic (collect-then-apply)
# ---------------------------------------------------------------------------
class TestReassignment:
    """Verify the collect-then-apply pattern doesn't skip items."""

    def test_no_skip(self):
        class FakeTx:
            def __init__(self, name, seq):
                self.name = name
                self._seq = seq

        named = {
            "geneA": [FakeTx("a1", "ACGTACGTACGT")],
            "geneB": [FakeTx("b1", "GGGGGGGG")],
            "unnamed": [
                FakeTx("u1", "ACGT"),  # too short ratio for geneA
                FakeTx("u2", "ACGTACGT"),  # matches geneA (8 > 0.5*12)
                FakeTx("u3", "GGGG"),  # too short ratio for geneB
            ],
        }

        reassigned = []
        for tx in named["unnamed"]:
            matched = False
            for target_gene, tx_list in named.items():
                if target_gene == "unnamed":
                    continue
                for tx2 in tx_list:
                    if tx._seq in tx2._seq and len(tx._seq) > 0.5 * len(tx2._seq):
                        reassigned.append((tx, target_gene))
                        matched = True
                        break
                if matched:
                    break

        for tx, target_gene in reassigned:
            named["unnamed"].remove(tx)
            named[target_gene].append(tx)

        assert len(named["unnamed"]) == 2
        assert len(named["geneA"]) == 2
        assert any(tx.name == "u2" for tx in named["geneA"])


# ---------------------------------------------------------------------------
# Integration: group_genes (requires test data)
# ---------------------------------------------------------------------------
class TestGroupGenes:
    def test_basic(self, tmp_path, data_dir, has_gtf_data):
        from coralsnake.genegroup import group_genes

        output = str(tmp_path / "output.tsv")
        consensus = str(tmp_path / "consensus.fa")
        group_genes(
            fa_file_list=[str(data_dir / "R64-1-1.fa")],
            gtf_file_list=[str(data_dir / "R64-1-1.release57.gtf")],
            out_file=output,
            consensus_fa=consensus,
            threads=2,
        )
        from pathlib import Path

        assert Path(output).exists()
        assert Path(output).stat().st_size > 0



"""Tests for coralsnake.gtf2tx – GTF/GFF parsing and transcript ranking."""

from coralsnake.gtf2tx import (
    parse_gff_annot,
    parse_gtf_annot,
    rank_transcript,
    sanitize_sequence_name,
)
from coralsnake.utils import Transcript


# ---------------------------------------------------------------------------
# GTF / GFF attribute parsing
# ---------------------------------------------------------------------------
class TestParseGtfAnnot:
    def test_basic(self):
        d = parse_gtf_annot('gene_id "ENSG001"; transcript_id "ENST001";')
        assert d["gene_id"] == "ENSG001"
        assert d["transcript_id"] == "ENST001"

    def test_duplicate_keys(self):
        d = parse_gtf_annot('tag "basic"; tag "MANE_Select";')
        assert "basic" in d["tag"]
        assert "MANE_Select" in d["tag"]

    def test_empty(self):
        d = parse_gtf_annot("")
        assert d == {}


class TestParseGffAnnot:
    def test_basic(self):
        d = parse_gff_annot("ID=exon1;Parent=mRNA1")
        assert d["ID"] == "exon1"
        assert d["Parent"] == "mRNA1"

    def test_empty(self):
        d = parse_gff_annot("")
        assert d == {}


# ---------------------------------------------------------------------------
# rank_transcript
# ---------------------------------------------------------------------------
class TestRankTranscript:
    def test_mane_select_highest(self):
        tx = Transcript()
        tx.priority = (0, 0)
        assert rank_transcript("T1", tx) == (0, 0)

    def test_ensembl_canonical(self):
        tx = Transcript()
        tx.priority = (0, 1)
        assert rank_transcript("T1", tx) == (0, 1)

    def test_dot_suffix(self):
        tx = Transcript()
        tx.priority = (10, 0)
        assert rank_transcript("AT1G01010.1", tx) == (2, 1)

    def test_dash_suffix(self):
        tx = Transcript()
        tx.priority = (10, 0)
        assert rank_transcript("LOC_Os01g01010-01", tx) == (3, 1)

    def test_fallback_length(self):
        tx = Transcript(
            exons={1: __import__("coralsnake.utils", fromlist=["Span"]).Span(0, 500)},
            strand="+",
        )
        tx.priority = (10, 0)
        r = rank_transcript("UNKNOWN", tx)
        assert r[0] == 4


# ---------------------------------------------------------------------------
# sanitize_sequence_name
# ---------------------------------------------------------------------------
class TestSanitizeSequenceName:
    def test_clean_name(self):
        assert sanitize_sequence_name("ENSG00000001") == "ENSG00000001"

    def test_special_chars(self):
        result = sanitize_sequence_name("gene (copy)")
        assert "(" not in result
        assert "_" in result

    def test_empty(self):
        assert sanitize_sequence_name("") == ""


# ---------------------------------------------------------------------------
# Integration: parse_file
# ---------------------------------------------------------------------------
class TestParseFile:
    def test_basic(self, tmp_path, data_dir, has_gtf_data):
        from pathlib import Path
        from coralsnake.gtf2tx import parse_file

        output = str(tmp_path / "output.tsv")
        parse_file(
            gtf_file=str(data_dir / "R64-1-1.release57.gtf"),
            fasta_file=str(data_dir / "R64-1-1.fa"),
            output_file=output,
            seq_file=None,
            sanitize=False,
            with_codon=False,
            with_genename=False,
            with_biotype=False,
            with_txpos=False,
            filter_biotype=None,
            seq_upper=True,
            line_length=0,
        )
        content = Path(output).read_text()
        lines = content.strip().split("\n")
        assert len(lines) > 1  # header + data

    def test_with_seq(self, tmp_path, data_dir, has_gtf_data):
        from pathlib import Path
        from coralsnake.gtf2tx import parse_file

        output = str(tmp_path / "output.tsv")
        seq_out = str(tmp_path / "seqs.fa")
        parse_file(
            gtf_file=str(data_dir / "R64-1-1.release57.gtf"),
            fasta_file=str(data_dir / "R64-1-1.fa"),
            output_file=output,
            seq_file=seq_out,
            sanitize=False,
            with_codon=False,
            with_genename=False,
            with_biotype=False,
            with_txpos=False,
            filter_biotype=None,
            seq_upper=True,
            line_length=0,
        )
        assert Path(seq_out).exists()
        seq_content = Path(seq_out).read_text()
        assert seq_content.startswith(">")



#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-08 20:32


import atexit
import multiprocessing as mp
import os
import random
import re
import shutil
import sys
import tempfile
import time
from collections import deque
from concurrent.futures import ProcessPoolExecutor, as_completed, wait, FIRST_COMPLETED
from contextlib import ExitStack
from functools import lru_cache

import pysam
from bwamem import BwaAligner, BwaIndexer, fastx_read, read_paired_fastx
from bwamem.libbwa import ffi, libbwa
from rich.progress import Progress, SpinnerColumn, TextColumn

from . import seqops
from .utils import (
    convert_file_realtime,
    format_duration,
    km_conversion,
    mk_conversion,
    reverse_complement,
)


@lru_cache(maxsize=100)
def _get_ref_cached(idx_ptr, rid, start, end):
    res_ptr = libbwa.bwa_fetch_seq(idx_ptr, rid, start, end)
    if res_ptr == ffi.NULL:
        return None
    res = ffi.string(res_ptr).decode()
    libbwa.free(res_ptr)
    return res


def _build_indices_with_progress(
    ref_file: str,
    index_base_dir: str,
    on_update,
    poll_interval: float = 0.1,
    ref_suffix: str = "",
):
    from concurrent.futures import ThreadPoolExecutor

    os.makedirs(index_base_dir, exist_ok=True)
    orig_fa = os.path.join(index_base_dir, f"ref{ref_suffix}.orig.fa")
    orig_prefix = os.path.splitext(orig_fa)[0]
    mk_fa = os.path.join(index_base_dir, f"ref{ref_suffix}.mk.fa")
    mk_prefix = os.path.splitext(mk_fa)[0]

    if ref_file is None:
        on_update("Done", 100.0, 100.0)
        return orig_fa, mk_prefix

    if os.path.abspath(ref_file) != os.path.abspath(orig_fa):
        shutil.copyfile(ref_file, orig_fa)

    indexer1 = BwaIndexer(verbose=2, capture_progress=True)
    indexer2 = BwaIndexer(verbose=2, capture_progress=True)

    def build_orig():
        indexer1.build_index(orig_fa, prefix=orig_prefix)

    def build_mk():
        indexer2.build_index(mk_fa, prefix=mk_prefix)

    def do_conversion():
        convert_file_realtime(ref_file, mk_fa, "AC", "GT")

    with ThreadPoolExecutor(max_workers=3) as ex:
        fut_orig = ex.submit(build_orig)
        fut_conv = ex.submit(do_conversion)
        while not fut_conv.done():
            p1 = 100.0 if fut_orig.done() else (indexer1.progress_percent or 0.0)
            on_update("Converting...", p1, 0.0)
            time.sleep(poll_interval)
        fut_conv.result()
        fut_mk = ex.submit(build_mk)
        while True:
            p1 = 100.0 if fut_orig.done() else (indexer1.progress_percent or 0.0)
            p2 = 100.0 if fut_mk.done() else (indexer2.progress_percent or 0.0)
            on_update("Indexing...", p1, p2)
            if fut_orig.done() and fut_mk.done():
                break
            time.sleep(poll_interval)
        fut_orig.result()
        fut_mk.result()

    on_update("Done", 100.0, 100.0)
    return orig_fa, mk_prefix


def _map_batch_worker(
    batch, paired, max_mismatches, min_alignment_length, min_mapping_ratio
):
    results = []
    if not paired:
        seqs = [item[1] for item in batch]
        conv1 = (
            seqops.batch_base_conversion(seqs, "AC", "GT")
            if _FORWARD_LIBRARY
            else seqops.batch_base_conversion(seqs, "GT", "AC")
        )
        conv2 = (
            seqops.batch_base_conversion(seqs, "GT", "AC")
            if _FORWARD_LIBRARY
            else seqops.batch_base_conversion(seqs, "AC", "GT")
        )
        for i, (name1, seq1, qua1) in enumerate(batch):
            ref_idx, mapping_result = run_mapping_se(
                seq1,
                qua1,
                conv1[i],
                conv2[i],
                max_mismatches,
                min_alignment_length,
                min_mapping_ratio,
            )
            results.append((ref_idx, mapping_result))
    else:
        seqs1 = [item[0][1] for item in batch]
        seqs2 = [item[1][1] for item in batch]
        rc_s1 = [seqops.reverse_complement(s) for s in seqs1]
        rc_s2 = [seqops.reverse_complement(s) for s in seqs2]
        if _FORWARD_LIBRARY:
            c1_r1, c1_r2 = (
                seqops.batch_base_conversion(seqs1, "AC", "GT"),
                seqops.batch_base_conversion(rc_s2, "AC", "GT"),
            )
            c2_r1, c2_r2 = (
                seqops.batch_base_conversion(rc_s1, "AC", "GT"),
                seqops.batch_base_conversion(seqs2, "AC", "GT"),
            )
        else:
            c1_r1, c1_r2 = (
                seqops.batch_base_conversion(rc_s1, "AC", "GT"),
                seqops.batch_base_conversion(seqs2, "AC", "GT"),
            )
            c2_r1, c2_r2 = (
                seqops.batch_base_conversion(seqs1, "AC", "GT"),
                seqops.batch_base_conversion(rc_s2, "AC", "GT"),
            )
        conv_data = (c1_r1, c1_r2, c2_r1, c2_r2)
        for i, ((name1, seq1, qua1), (name2, seq2, qua2)) in enumerate(batch):
            ref_idx, mapping_result = run_mapping_pe(
                seq1,
                seq2,
                qua1,
                qua2,
                i,
                conv_data,
                rc_s1[i],
                rc_s2[i],
                max_mismatches,
                min_alignment_length,
                min_mapping_ratio,
            )
            results.append((ref_idx, mapping_result))
    return results


def _check_reference_length(aligner, min_length=100):
    for i in range(aligner.index.bns.n_seqs):
        if aligner.index.bns.anns[i].len >= min_length:
            return True
    return False


def run_mapping_se(
    seq1, qua1, s1_c1, s1_c2, max_mismatches, min_alignment_length, min_mapping_ratio
):
    for ref_idx in range(len(_ALIGNERS_MK)):
        if not _check_reference_length(_ALIGNERS_MK[ref_idx]):
            continue
        mapped = []
        oris = [1, 2] if _ORIENTATION_FILTER is None else [_ORIENTATION_FILTER]
        for orientation in oris:
            s1_c = s1_c1 if orientation == 1 else s1_c2
            hits = _ALIGNERS_MK[ref_idx].align(s1_c, min_mapq=1)
            if not hits:
                continue
            for h in hits:
                if (h[2] - h[1]) < min_alignment_length:
                    continue
                is_rev = h[3] == -1
                s = seqops.reverse_complement(seq1) if is_rev else seq1
                rid = h[10] if len(h) > 10 else _RID_MAPS[ref_idx].get(h[0], -1)
                if rid < 0:
                    continue
                ref = _get_ref_cached(
                    _ALIGNERS_ORIG[ref_idx].index, rid, h[1], h[2] + 100
                )
                res = seqops.score_and_tag(h[7], s, ref, (orientation == 1) ^ is_rev)
                if (
                    not res
                    or res[0] < -500
                    or res[1] > max_mismatches // 2
                    or (h[5] - h[4]) < len(seq1) * min_mapping_ratio
                ):
                    continue
                mapped.append(
                    [
                        is_rev,
                        h[0],
                        h[1],
                        score_to_mapq(res[0]),
                        h[7],
                        res[0],
                        res[2],
                        orientation,
                        res[3],
                        res[4],
                        res[5],
                        res[6],
                        res[7],
                        res[8],
                        rid,
                    ]
                )
        if mapped:
            mapped.sort(key=lambda x: x[5], reverse=True)
            return (ref_idx, mapped)
    return (None, [])


def run_mapping_pe(
    seq1,
    seq2,
    qua1,
    qua2,
    idx,
    conv_data,
    rc_s1,
    rc_s2,
    max_mismatches,
    min_alignment_length,
    min_mapping_ratio,
):
    q1_l, q2_l = len(seq1), len(seq2)
    c1_r1, c1_r2, c2_r1, c2_r2 = conv_data
    all_res = []
    for ref_idx in range(len(_ALIGNERS_MK)):
        if not _check_reference_length(_ALIGNERS_MK[ref_idx]):
            continue
        mapped = []
        oris = [1, 2] if _ORIENTATION_FILTER is None else [_ORIENTATION_FILTER]
        for orientation in oris:
            is_o1 = orientation == 1
            s1_c, s2_c = (c1_r1[idx], c1_r2[idx]) if is_o1 else (c2_r1[idx], c2_r2[idx])
            hits = _ALIGNERS_MK[ref_idx].align(s1_c, s2_c, min_mapq=1)
            if not hits:
                continue
            for h1, h2, is_p, isize in hits:
                res1 = res2 = None
                if h1 and (h1[2] - h1[1]) >= min_alignment_length:
                    r1_rev = (
                        (h1[3] == -1 if _FORWARD_LIBRARY else h1[3] == 1)
                        if is_o1
                        else (h1[3] == 1 if _FORWARD_LIBRARY else h1[3] == -1)
                    )
                    s1_cur = rc_s1 if r1_rev else seq1
                    rid1 = h1[10] if len(h1) > 10 else _RID_MAPS[ref_idx].get(h1[0], -1)
                    if rid1 >= 0:
                        ref1 = _get_ref_cached(
                            _ALIGNERS_ORIG[ref_idx].index, rid1, h1[1], h1[2] + 100
                        )
                        res1 = seqops.score_and_tag(
                            h1[7],
                            s1_cur,
                            ref1,
                            (is_o1 if _FORWARD_LIBRARY else (not is_o1)) ^ r1_rev,
                        )
                if h2 and (h2[2] - h2[1]) >= min_alignment_length:
                    r2_rev = (
                        (h2[3] == 1 if _FORWARD_LIBRARY else h2[3] == -1)
                        if is_o1
                        else (h2[3] == -1 if _FORWARD_LIBRARY else h2[3] == 1)
                    )
                    s2_cur = rc_s2 if r2_rev else seq2
                    rid2 = h2[10] if len(h2) > 10 else _RID_MAPS[ref_idx].get(h2[0], -1)
                    if rid2 >= 0:
                        ref2 = _get_ref_cached(
                            _ALIGNERS_ORIG[ref_idx].index, rid2, h2[1], h2[2] + 100
                        )
                        res2 = seqops.score_and_tag(
                            h2[7],
                            s2_cur,
                            ref2,
                            (not (is_o1 if _FORWARD_LIBRARY else (not is_o1))) ^ r2_rev,
                        )
                if res1 or res2:
                    if (res1 and res1[0] < -500) or (res2 and res2[0] < -500):
                        continue
                    if (
                        (res1[1] if res1 else 0) + (res2[1] if res2 else 0)
                    ) > max_mismatches:
                        continue
                    if h1 and (h1[5] - h1[4]) < q1_l * min_mapping_ratio:
                        continue
                    if h2 and (h2[5] - h2[4]) < q2_l * min_mapping_ratio:
                        continue
                    mq = score_to_mapq(
                        min(res1[0] if res1 else 99, res2[0] if res2 else 99)
                    )
                    m1 = (
                        [
                            h1[3] == -1,
                            h1[0],
                            h1[1],
                            mq,
                            h1[7],
                            h2[0] if h2 else "*",
                            h2[1] if h2 else 0,
                            (isize if h1[1] <= h2[1] else -isize) if h2 else 0,
                            res1[0],
                            res1[2],
                            orientation,
                            res1[3],
                            res1[4],
                            res1[5],
                            res1[6],
                            res1[7],
                            res1[8],
                            rid1,
                            bool(is_p),
                            bool(h2 and h2[3] == -1),
                            bool(not h2),
                        ]
                        if res1
                        else None
                    )
                    m2 = (
                        [
                            h2[3] == -1,
                            h2[0],
                            h2[1],
                            mq,
                            h2[7],
                            h1[0] if h1 else "*",
                            h1[1] if h1 else 0,
                            (isize if h2[1] <= h1[1] else -isize) if h1 else 0,
                            res2[0],
                            res2[2],
                            orientation,
                            res2[3],
                            res2[4],
                            res2[5],
                            res2[6],
                            res2[7],
                            res2[8],
                            rid2,
                            bool(is_p),
                            bool(h1 and h1[3] == -1),
                            bool(not h1),
                        ]
                        if res2
                        else None
                    )
                    mapped.append(
                        [(res1[0] if res1 else 0) + (res2[0] if res2 else 0), m1, m2]
                    )
        if mapped:
            mapped.sort(key=lambda x: x[0], reverse=True)
            if any(r[1] and r[2] for r in mapped):
                return (ref_idx, mapped)
            all_res.append((ref_idx, mapped))
        else:
            all_res.append((ref_idx, []))
    for r_idx, rs in all_res:
        if rs:
            return (r_idx, rs)
    return (None, [])


_ALIGNERS_ORIG = []
_ALIGNERS_MK = []
_ORIENTATION_FILTER = None
_FORWARD_LIBRARY = None
_RID_MAPS = []


def _init_worker(ref_indices, orientation_filter, forward_library):
    global \
        _ALIGNERS_ORIG, \
        _ALIGNERS_MK, \
        _ORIENTATION_FILTER, \
        _FORWARD_LIBRARY, \
        _RID_MAPS
    _ORIENTATION_FILTER, _FORWARD_LIBRARY = orientation_filter, forward_library
    _ALIGNERS_ORIG = []
    _ALIGNERS_MK = []
    _RID_MAPS = []
    for o_fa, mk_pre, _ in ref_indices:
        opts = {
            "min_seed_len": 14,
            "max_occ": 1000,
            "softclip_supplementary": True,
            "mark_secondary": True,
            "clip_penalties": (6, 6),
            "unpaired_penalty": 24,
            "min_score": 20,
            "insert_model": (80, 60, 450),
        }
        a_orig = BwaAligner(os.path.splitext(o_fa)[0], **opts)
        a_mk = BwaAligner(mk_pre, **opts)
        _ALIGNERS_ORIG.append(a_orig)
        _ALIGNERS_MK.append(a_mk)
        _RID_MAPS.append(
            {
                ffi.string(a_orig.index.bns.anns[i].name).decode(): i
                for i in range(a_orig.index.bns.n_seqs)
            }
        )


def _build_and_check_indices(ref_files, ref_indices, index_dirs, index_only):
    with Progress(
        SpinnerColumn(),
        TextColumn("[bold green]Index[/bold green]"),
        TextColumn("{task.description}"),
        TextColumn(
            "| [yellow]{task.fields[status]}[/yellow] | [cyan]{task.fields[progress]}[/cyan]"
        ),
        transient=False,
    ) as progress:
        task = progress.add_task(
            f"Indices...", total=None, status="Checking", progress="0%"
        )
        for i in range(len(ref_indices)):
            ref_suffix = ref_indices[i][2]
            d = index_dirs[i] if len(index_dirs) == len(ref_indices) else index_dirs[0]
            o_fa, mk_pre = (
                os.path.join(d, f"ref{ref_suffix}.orig.fa"),
                os.path.join(d, f"ref{ref_suffix}.mk"),
            )
            if index_only or not (
                os.path.exists(o_fa)
                and all(
                    os.path.exists(mk_pre + e)
                    for e in [".amb", ".ann", ".bwt", ".pac", ".sa"]
                )
            ):

                def on_up(s, p1, p2):
                    progress.update(
                        task,
                        status=f"Ref {i + 1}: {s}",
                        progress=f"{(p1 + p2) / 2:.1f}%",
                    )

                idx0, idx_mk = _build_indices_with_progress(
                    ref_files[i] if i < len(ref_files) else None,
                    d,
                    on_up,
                    ref_suffix=ref_suffix,
                )
                ref_indices[i] = (idx0, idx_mk, ref_suffix)
            else:
                ref_indices[i] = (o_fa, mk_pre, ref_suffix)
        progress.update(
            task, description="✓ Indices ready", status="Done", progress="100%"
        )
        return ref_indices


def score_to_mapq(score):
    return max(0, min(60, score))


def create_bam_record(
    header, map_data, name, seq, qual, is_secondary, global_rid_map, read_id
):
    a = pysam.AlignedSegment(header=header)
    a.query_name = name
    a.is_paired = True
    a.is_read1 = read_id == 1
    a.is_read2 = read_id == 2
    a.is_secondary = is_secondary
    a.reference_id = global_rid_map.get(map_data[1], -1)
    a.reference_start, a.mapping_quality, a.cigarstring = (
        map_data[2],
        map_data[3],
        map_data[4],
    )
    a.is_reverse = bool(map_data[0])
    if len(map_data) > 18:
        a.is_proper_pair = bool(map_data[18])
        a.mate_is_reverse = bool(map_data[19])
        a.mate_is_unmapped = bool(map_data[20])
        a.next_reference_id = global_rid_map.get(map_data[5], -1)
        a.next_reference_start = map_data[6] if not a.mate_is_unmapped else 0
        a.template_length = map_data[7]
    else:
        a.is_paired = False
    if a.is_reverse:
        a.query_sequence = reverse_complement(seq)
        a.query_qualities = pysam.qualitystring_to_array(qual[::-1])
    else:
        a.query_sequence = seq
        a.query_qualities = pysam.qualitystring_to_array(qual)
    off = 8 if len(map_data) > 18 else 5
    tags = [
        ("AS", int(map_data[off])),
        ("MD", str(map_data[off + 1])),
        ("ST", int(map_data[off + 2])),
    ]
    for i, tag in enumerate(["Yf", "Zf", "Yc", "Zc", "NS", "NC"]):
        tags.append((tag, int(map_data[off + 3 + i])))
    a.tags = tags
    return a


def create_unmapped_record(header, name, seq, qual, flag):
    a = pysam.AlignedSegment(header=header)
    a.query_name, a.flag, a.reference_id, a.reference_start, a.mapping_quality = (
        name,
        flag,
        -1,
        -1,
        0,
    )
    a.query_sequence, a.query_qualities = seq, pysam.qualitystring_to_array(qual)
    return a


def _setup_output_bams(
    output_files, unmap_file, unified_header, ref_headers, stack, threads=1
):
    if len(output_files) > 1:
        bam_outs = {
            i: stack.enter_context(
                pysam.AlignmentFile(f, "wb", header=ref_headers[i], threads=threads)
            )
            for i, f in enumerate(output_files)
        }
        default_bam = bam_outs[len(output_files) - 1]
    else:
        default_bam = stack.enter_context(
            pysam.AlignmentFile(
                output_files[0], "wb", header=unified_header, threads=threads
            )
        )
        bam_outs = None
    bam_unmap = (
        stack.enter_context(
            pysam.AlignmentFile(
                unmap_file, "wb", header=unified_header, threads=threads
            )
        )
        if unmap_file
        else default_bam
    )
    return bam_outs, bam_unmap, default_bam


def map_file(
    r1_file,
    r2_file,
    ref_files,
    output_files,
    unmap_file=None,
    forward_library=True,
    max_mismatches=10,
    threads=8,
    min_alignment_length=20,
    min_mapping_ratio=0.5,
    index_dir=None,
    index_only=False,
    batch_size=2000,
    orientation_filter=None,
):
    global _ORIENTATION_FILTER, _FORWARD_LIBRARY
    _ORIENTATION_FILTER, _FORWARD_LIBRARY = orientation_filter, forward_library
    ref_indices = [
        (None, None, str(i + 1) if len(ref_files) > 1 else "")
        for i in range(len(ref_files))
    ]
    i_dirs = (
        [index_dir]
        if index_dir and not isinstance(index_dir, (list, tuple))
        else (index_dir or [os.path.dirname(os.path.abspath(f)) for f in ref_files])
    )
    ref_indices = _build_and_check_indices(ref_files, ref_indices, i_dirs, index_only)
    if index_only:
        return
    mp_ctx = (
        mp.get_context("fork") if sys.platform != "win32" else mp.get_context("spawn")
    )
    stack = ExitStack()
    with stack:
        ref_headers = []
        all_sq = []
        seen_sq = set()
        for o_fa, _, _ in ref_indices:
            with pysam.FastaFile(o_fa) as fa:
                h = {
                    "HD": {"VN": "1.6", "SO": "unsorted"},
                    "SQ": [
                        {"LN": fa.get_reference_length(n), "SN": n}
                        for n in fa.references
                    ],
                }
                ref_headers.append(h)
                [
                    (all_sq.append(sq), seen_sq.add(sq["SN"]))
                    for sq in h["SQ"]
                    if sq["SN"] not in seen_sq
                ]
        unified_h = {"HD": {"VN": "1.6", "SO": "unsorted"}, "SQ": all_sq}
        global_rid_map = {sq["SN"]: i for i, sq in enumerate(all_sq)}
        global_rid_map["*"] = -1
        bam_outs, bam_unmap, def_bam = _setup_output_bams(
            output_files, unmap_file, unified_h, ref_headers, stack, threads=2
        )
        p_tot = p_map = 0
        s_time = time.time()
        with ProcessPoolExecutor(
            max_workers=max(1, threads),
            mp_context=mp_ctx,
            initializer=_init_worker,
            initargs=(ref_indices, orientation_filter, forward_library),
        ) as ex:
            futures = deque()
            batches = deque()
            it_reads = (
                read_paired_fastx(r1_file, r2_file) if r2_file else fastx_read(r1_file)
            )
            batch = []
            with Progress(
                SpinnerColumn(),
                TextColumn("[bold blue]Map[/bold blue]"),
                TextColumn("{task.fields[mapped]} / {task.fields[total_reads]} reads"),
                TextColumn("({task.fields[elapsed]})"),
                transient=False,
            ) as prg:
                task = prg.add_task(
                    "Mapping", total=None, mapped="0", total_reads="0", elapsed="0.00s"
                )

                def flush_done():
                    nonlocal p_tot, p_map
                    while futures and futures[0].done():
                        res = futures.popleft().result()
                        orig_batch = batches.popleft()
                        for i, (ridx, m_res) in enumerate(res):
                            p_tot += 1
                            info = orig_batch[i]
                            if ridx is not None and m_res:
                                p_map += 1
                                target = bam_outs[ridx] if bam_outs else def_bam
                                for j, item in enumerate(m_res):
                                    if r2_file:
                                        if item[1]:
                                            target.write(
                                                create_bam_record(
                                                    target.header,
                                                    item[1],
                                                    info[0][0],
                                                    info[0][1],
                                                    info[0][2],
                                                    j > 0,
                                                    global_rid_map,
                                                    1,
                                                )
                                            )
                                        if item[2]:
                                            target.write(
                                                create_bam_record(
                                                    target.header,
                                                    item[2],
                                                    info[1][0],
                                                    info[1][1],
                                                    info[1][2],
                                                    j > 0,
                                                    global_rid_map,
                                                    2,
                                                )
                                            )
                                    else:
                                        target.write(
                                            create_bam_record(
                                                target.header,
                                                item,
                                                info[0],
                                                info[1],
                                                info[2],
                                                j > 0,
                                                global_rid_map,
                                                0,
                                            )
                                        )
                            else:
                                if r2_file:
                                    bam_unmap.write(
                                        create_unmapped_record(
                                            bam_unmap.header,
                                            info[0][0],
                                            info[0][1],
                                            info[0][2],
                                            77,
                                        )
                                    )
                                    bam_unmap.write(
                                        create_unmapped_record(
                                            bam_unmap.header,
                                            info[1][0],
                                            info[1][1],
                                            info[1][2],
                                            141,
                                        )
                                    )
                                else:
                                    bam_unmap.write(
                                        create_unmapped_record(
                                            bam_unmap.header,
                                            info[0],
                                            info[1],
                                            info[2],
                                            4,
                                        )
                                    )
                        prg.update(
                            task,
                            mapped=f"{p_map:,}",
                            total_reads=f"{p_tot:,}",
                            elapsed=format_duration(time.time() - s_time),
                        )

                for reads in it_reads:
                    if r2_file:
                        if reads[0].name.split()[0].rstrip("/1").rstrip("/2") != reads[
                            1
                        ].name.split()[0].rstrip("/1").rstrip("/2"):
                            raise ValueError(
                                f"Order mismatch: {reads[0].name} vs {reads[1].name}"
                            )
                        batch.append(
                            (
                                (reads[0].name, reads[0].sequence, reads[0].quality),
                                (reads[1].name, reads[1].sequence, reads[1].quality),
                            )
                        )
                    else:
                        batch.append((reads.name, reads.sequence, reads.quality))
                    if len(batch) >= batch_size:
                        batches.append(list(batch))
                        futures.append(
                            ex.submit(
                                _map_batch_worker,
                                batches[-1],
                                r2_file is not None,
                                max_mismatches,
                                min_alignment_length,
                                min_mapping_ratio,
                            )
                        )
                        batch.clear()
                        while len(futures) >= threads * 2:
                            wait(futures, return_when=FIRST_COMPLETED)
                            flush_done()
                if batch:
                    batches.append(list(batch))
                    futures.append(
                        ex.submit(
                            _map_batch_worker,
                            batches[-1],
                            r2_file is not None,
                            max_mismatches,
                            min_alignment_length,
                            min_mapping_ratio,
                        )
                    )
                while futures:
                    wait(futures, return_when=FIRST_COMPLETED)
                    flush_done()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-07-04 23:20


import re
from collections import defaultdict

import numpy as np
import rich.progress
from pyfamsa import Aligner, Sequence
from pysam import FastaFile
from scipy.cluster.hierarchy import fcluster, linkage
from scipy.spatial.distance import pdist

from .gtf2tx import read_gtf, top_transcript
from .utils import get_logger

LOGGER = get_logger(__name__)


def run_msa(names, seqs, threads=8):
    aligner = Aligner(guide_tree="upgma", threads=threads)
    seqs = [Sequence(n.encode(), s.encode()) for n, s in zip(names, seqs)]
    # tree = aligner.build_tree(seqs)
    # print(tree.dumps())
    msa = aligner.align(seqs)
    return msa


def msa_to_array(msa, mask_ratio=0.5):
    arr = np.array([list(sequence.sequence) for sequence in msa])
    return arr


def cluster_sequences(alignment_array, threshold):
    # Calculate pairwise Hamming distances using scipy's pdist function
    def hamming(u, v, **kwargs):
        # return np.sum(u != v) / len(u)
        # but do not calculate the distance of positions where both sequence is a gap
        # note that if one sequence is a gap and the other is not, the distance is 1
        mask = (u == 45) & (v == 45)
        u1 = u[~mask]
        v1 = v[~mask]
        return np.sum(u1 != v1) / len(u1)

    distance_matrix = pdist(alignment_array, metric=hamming)

    if len(distance_matrix) == 0:
        return np.zeros(alignment_array.shape[0], dtype=int)

    # Perform hierarchical clustering using the 'average' method
    Z = linkage(distance_matrix, method="average")

    # Form clusters with a maximum distance threshold
    raw_clusters = fcluster(Z, t=threshold, criterion="distance")

    # Use NumPy to count and sort clusters by size
    unique, counts = np.unique(raw_clusters, return_counts=True)
    sorted_clusters = unique[np.argsort(-counts)]

    # Map original cluster IDs to new ordered IDs
    cluster_map = {cluster_id: i + 1 for i, cluster_id in enumerate(sorted_clusters)}
    ordered_clusters = np.vectorize(cluster_map.get)(raw_clusters)

    return ordered_clusters


def consensus_sequence(arr):
    # find the most common character in each column, and join to get the consensus sequence
    consensus = ""
    for i in range(arr.shape[1]):
        col = arr[:, i]
        unique, counts = np.unique(col, return_counts=True)
        max_count_index = np.argmax(counts)
        consensus += chr(unique[max_count_index])
    return consensus


def get_position_mapping_from_aligned_array(aligned_array):
    def map_positions(aligned_sequence):
        gap_positions = np.where(aligned_sequence == 45)[0]  # Find positions of '-'
        all_positions = np.arange(len(aligned_sequence))
        non_gap_positions = np.setdiff1d(all_positions, gap_positions)

        original_positions = []
        aligned_positions = []

        if len(non_gap_positions) == 0:
            return original_positions, aligned_positions

        start_orig = 0
        start_aligned = non_gap_positions[0]

        for i in range(1, len(non_gap_positions)):
            if non_gap_positions[i] != non_gap_positions[i - 1] + 1:
                end_orig = start_orig + (non_gap_positions[i - 1] - start_aligned) + 1
                end_aligned = non_gap_positions[i - 1] + 1

                original_positions.append((start_orig, end_orig))
                aligned_positions.append((start_aligned, end_aligned))

                start_orig = end_orig
                start_aligned = non_gap_positions[i]

        end_orig = start_orig + (non_gap_positions[-1] - start_aligned) + 1
        end_aligned = non_gap_positions[-1] + 1

        original_positions.append((start_orig, end_orig))
        aligned_positions.append((start_aligned, end_aligned))

        return list(zip(original_positions, aligned_positions))

    for i in range(aligned_array.shape[0]):
        aligned_sequence = aligned_array[i]
        mappings = map_positions(aligned_sequence)
        yield mappings


def rename_snRNA(gene_name):
    # This a temporary fix for snRNA naming
    # RNU2-1 -> U2
    # RNU2-27P -> U2
    # RNU3-2 -> U3
    # RNU5E-10P -> RN5E
    # RNU5A-3P -> RN5E
    # RNU4ATAC7 -> U4ATAC -> U4atac
    # RNU6ATAC7 -> U6ATAC -> U6atac
    # RNVU1-1 -> U1V
    # .., etc
    # for drosophila, snRNA:U2:38ABb -> U2
    if gene_name.startswith("snRNA:"):
        gene_name = gene_name.split(":")[1]
        return gene_name
    patterns = [
        (r"RNU(\d+)(?:-\d+.*)?$", r"U\1"),
        (r"RNU(\d+[A-Z]+)(?:-\d+.*)?$", r"U\1"),
        (r"RNU(\d+)ATAC.*", r"U\1atac"),
        (r"Rnu(\d+)(?:-\d+.*)?$", r"U\1"),
        (r"Rnu(\d+[a-z]+)(?:-\d+.*)?$", r"U\1"),
        (r"Rnu(\d+)atac.*", r"U\1atac"),
        (r"RNVU(\d+)-\d+.*", r"U\1V"),
        (r"[Uu](\d+)ATAC$", r"U\1atac"),
        (r"u(\d+)$", r"U\1"),
        (r"^snrnau(.*)$", r"U\1"),
    ]

    # Apply each pattern and replacement
    for pattern, replacement in patterns:
        gene_name = re.sub(pattern, replacement, gene_name)
    return gene_name


def rename_snoRNA(gene_name):
    # replace rule:
    # replace U3 with SNORD3A
    # replace U3b1 with SNORD3B-1
    # replace U8 with SNORD118
    # replace small nucleolar RNA ACA64 with SNORA64
    # replace Z6
    if gene_name.upper().startswith("ACA"):
        match = re.match(r"ACA(\d+)([A-Z]?)-?(\d*)", gene_name.upper())
        if match:
            number, letter, copy = match.groups()
            # ignore copy number
            return f"SNORA{number}{letter}"
    elif gene_name.upper().startswith("U8"):
        return "SNORD118"
    elif gene_name.upper().startswith("U"):
        # match all other U RNA (WARNING: all snoRD? are there some SNORA?)
        match = re.match(r"U(\d+)([A-Z]?)-?(\d*)", gene_name.upper())
        if match:
            number, letter, copy = match.groups()
            if letter == "":
                letter = "A"
            return f"SNORD{number}{letter}"
    elif gene_name.upper() == "Z6":
        return "Z6"
    return gene_name


def group_genes(
    fa_file_list,
    gtf_file_list,
    out_file=None,
    consensus_fa=None,
    gene_name_regex=None,
    gene_biotype_list=None,
    gene_length_limit=300,
    cluster_threshold=0.1,
    threads=8,
):
    # by default miRNA, tRNA, snoRNA, ...
    if gene_biotype_list is None:
        gene_biotype_list = [
            "tRNA",
            "rRNA",
            "Mt_rRNA",
            "Mt_tRNA",
            "ribozyme",
            "vault_RNA",
            # "7SL_RNA", --> scRNA
            # "7SK_RNA", --> snRNA
            "Y_RNA",
            "miRNA",
            "scaRNA",
            "snoRNA",
            "snRNA",
            "scRNA",
            "ncRNA",
            "misc_RNA",
            "vault_RNA_pseudogene",
            # "7SL_pseudogene",--> scRNA
            # "7SK_pseudogene", --> snRNA
            # "processed_pseudogene",
            "pseudogene",
        ]
    else:
        gene_biotype_list = gene_biotype_list.split(",")

    chrom_to_fa = {}
    LOGGER.info("Loading fasta files")
    for fa_file in fa_file_list:
        fasta = FastaFile(fa_file)
        chrom_to_fa.update({chrom: fasta for chrom in fasta.references})

    LOGGER.info("Loading annotations from gtf file")
    gene_dict_by_name = defaultdict(lambda: defaultdict(list))
    for gtf_file in gtf_file_list:
        gene_dict = read_gtf(
            gtf_file,
            is_gff=gtf_file.endswith("gff") or gtf_file.endswith("gff3"),
            keep_annotation=True,
        )
        for transcript in top_transcript(gene_dict):
            if not transcript.gene_name or not transcript.transcript_biotype:
                continue
            gene_name = transcript.gene_name
            gene_product = transcript._annotations.get("product", "")

            # This is a patch to fix misc_RNA gene names
            # eg, 4       ensembl gene    88330176        88330278        .       -       .       gene_id "ENSG00000207480"; gene_version "1"; gene_name "Y_RNA"; gene_source "ensembl"; gene_biotype "misc_RNA";
            if transcript.transcript_biotype == "misc_RNA":
                # --- Step 1: Prioritize matching specific pseudogene patterns ---
                # Vault RNA pseudogene (e.g., "VTRNA3-1P")
                if gene_name.startswith("VTRNA") and "P" in gene_name:
                    transcript.transcript_biotype = "Vault_RNA_pseudogene"

                # 7SL pseudogene (e.g., "RN7SL521P")
                elif gene_name.startswith("RN7SL") and gene_name.endswith("P"):
                    transcript.transcript_biotype = "scRNA"  # "scRNA_pseudogene"

                # 7SK pseudogene (e.g., "RN7SKP27")
                elif gene_name.startswith("RN7SKP"):
                    transcript.transcript_biotype = "snRNA"  # "snRNA_pseudogene"

                # Y_RNA pseudogene (e.g., "RNY4P28", "RNY1P4")
                elif gene_name.startswith("RNY") and "P" in gene_name:
                    transcript.transcript_biotype = "Y_RNA_pseudogene"

                # --- Step 2: Match the broader functional RNA patterns ---

                # Functional 7SL RNA (e.g., "Metazoa_SRP", "RN7SL2", "7SL")
                elif gene_name.startswith("RN7SL") or gene_name in [
                    "Metazoa_SRP",
                    "7SL",
                ]:
                    transcript.transcript_biotype = "scRNA"

                # Functional Y_RNA (e.g., "Y_RNA", "RNY3", "RNY4")
                elif gene_name.startswith("RNY") or gene_name == "Y_RNA":
                    transcript.transcript_biotype = "Y_RNA"

                # Functional 7SK RNA (now checks for "RN7SK" or "7SK")
                elif gene_name in ["RN7SK", "7SK"]:
                    transcript.transcript_biotype = "snRNA"

                # Functional Vault RNA (so far, only "Vault" is observed)
                elif gene_name == "Vault":
                    transcript.transcript_biotype = "vault_RNA"

                # small Cajal body-specific RNA
                elif gene_name.startswith("SCARNA") or (
                    hasattr(transcript, "product")
                    and "small cajal body" in gene_product.lower()
                ):
                    transcript.transcript_biotype = "scaRNA"

            tx_biotype = transcript.transcript_biotype

            # # This is a temporary fix for misc_RNA naming
            # if tx_biotype == "misc_RNA":
            #     # for mouse reference, there are too many unknown genes
            #     if re.match(r"^Gm\d+$", gene_name):
            #         continue

            # This a temporary fix for snRNA naming
            gene_name = rename_snRNA(gene_name)
            if transcript.transcript_biotype == "snRNA":
                if gene_name.upper().startswith("7SK"):
                    gene_name = "7SK"
                elif gene_name.upper().startswith("SNAR-"):
                    gene_name = re.sub(r"[0-9-]+$", "", gene_name)
                elif gene_name.upper().startswith("U"):
                    pass
                elif (
                    "small nuclear RNA" in gene_product
                    or "spliceosomal RNA" in gene_product
                ):
                    # eg:
                    # U6 spliceosomal RNA
                    # U7 small nuclear RNA
                    gene_name = (
                        gene_product.replace("small nuclear RNA", "")
                        .replace("spliceosomal RNA", "")
                        .strip()
                        .split()[0]
                    )
                    gene_name = rename_snRNA(gene_name)
                else:
                    gene_name = "Ux"

            # This is a temporary fix for the naming of the Y_RNA genes
            if tx_biotype == "Y_RNA":
                if gene_name.upper() == "Y_RNA":
                    gene_name = "RNYx"

            # This is a temporary fix for the naming of the snoRNA genes
            # eg, SNORD115-1 and SNORD115-2 are the same gene
            if tx_biotype == "snoRNA":
                if gene_name.upper().startswith("SNOR"):
                    gene_name = gene_name.rsplit("-")[0]
                    # todo:
                    # SNORD109A and SNORD109B are the same gene
                    # SNORD62A and SNORD62B are the same gene
                elif gene_name.upper().startswith("U") or gene_name.upper().startswith(
                    "RNU"
                ):
                    gene_name = rename_snoRNA(gene_name)
                elif "small nucleolar" in gene_product:
                    # eg:
                    # small nucleolar RNA SNORD51
                    # small nucleolar SNORD12/SNORD106
                    _gene_name = (
                        gene_product.replace("small nucleolar RNA", "")
                        .replace("small nucleolar", "")
                        .strip()
                        .split()[0]
                    )
                    if (
                        gene_name.startswith("Gm")
                        or gene_name.startswith("AF")
                        or gene_name.startswith("DQ")
                    ):
                        gene_name = _gene_name.replace("SNORD", "Snord").replace(
                            "SNORA", "Snora"
                        )
                    elif not _gene_name.upper().startswith("SNOR"):
                        gene_name = rename_snoRNA(_gene_name)
                        if not gene_name.startswith("SNOR"):
                            gene_name = "SNORx"
                    else:
                        gene_name = _gene_name
                else:
                    gene_name = "SNORx"

            # This is a temporary fix for the naming of the scaRNA genes
            if tx_biotype == "scaRNA":
                if (
                    "small Cajal body-specific RNA" in gene_product
                    and not gene_name.upper().startswith("SCARNA")
                ):
                    gene_name = (
                        "SCARNA"
                        + gene_product.replace("small Cajal body-specific RNA", "")
                        .strip()
                        .split()[0]
                    )

            # This is a temporary fix for the naming of the miRNA genes
            if tx_biotype == "miRNA":
                if gene_name.upper().startswith("MIR"):
                    gene_name = gene_name.rsplit("-")[0]
                    # todo: MIR4436B1 and MIR4436B2 are the same gene
                    # patch: Mir3535 is snoRNA
                    if gene_name == "Mir3535":
                        tx_biotype = "snoRNA"
                        gene_name = "SNORx"
                elif "LET" in gene_name.upper() or "MIR" in gene_name.upper():
                    pass
                else:
                    gene_name = "MIRx"

            # This is a temporary fix for the naming of the tRNA genes
            if "_tRNA-" in gene_name:
                gene_name = "tRNA-" + gene_name.split("_tRNA-", 1)[1]
            if tx_biotype == "tRNA" and not gene_name.startswith("tRNA"):
                if gene_name == "":
                    gene_name = "unknown"
                gene_name = "tRNA-" + gene_name
            # Caenorhabditis_elegans_tRNA-Lys-CTT-1-7
            # remove the number suffix "-1-7"
            # loop from right to left and stop until reach a non-digit character
            # rmove digits and "-"
            if gene_name.startswith("tRNA-"):
                # gene_name = re.sub(r"-[0-9-]+$", "", gene_name)
                gene_name = re.sub(r"[0-9-]+$", "", gene_name)

            gene_dict_by_name[tx_biotype][gene_name].append(transcript)

    # sort the 1st level and 2nd level of gene_dict_by_name
    gene_dict_by_name = {
        k: dict(sorted(v.items(), key=lambda item: item[0]))
        for k, v in sorted(gene_dict_by_name.items(), key=lambda item: item[0])
    }

    # adjust gene dict, rule:
    # within each biotype, if the gene_name is one of ["SNORx", "MIRx", "Ux", "RNYx"]
    # iter the transcript list of this gene_name, and compare with other genes
    # 1. get the seq for each transcript (storge into tx._seq by default)
    tx_to_unname = {"snoRNA": "SNORx", "miRNA": "MIRx", "snRNA": "Ux", "Y_RNA": "RNYx"}
    for tx_biotype, tx_dict in rich.progress.track(
        gene_dict_by_name.items(),
        description="Preprocessing genes...",
    ):
        if all(not (x == tx_biotype and y in tx_dict) for x, y in tx_to_unname.items()):
            continue
        for gene_name, tx_list in rich.progress.track(
            tx_dict.items(),
            description=f"Preprocessing genes ({tx_biotype})...",
        ):
            for tx in tx_list:
                tx._seq = tx.get_seq(chrom_to_fa[tx.chrom])
        # 2. for each biotype, loop tx in gene_name in ["SNORx", "MIRx", "Ux", "RNYx"]
        unnamed_gene = tx_to_unname[tx_biotype]
        reassigned = []
        for tx in tx_dict[unnamed_gene]:
            matched = False
            for target_gene, tx_list in tx_dict.items():
                if target_gene == unnamed_gene:
                    continue
                for tx2 in tx_list:
                    if (tx._seq in tx2._seq and len(tx._seq) > 0.5 * len(tx2._seq)) or (
                        tx2._seq in tx._seq and len(tx2._seq) > 0.9 * len(tx._seq)
                    ):
                        reassigned.append((tx, target_gene))
                        matched = True
                        break
                if matched:
                    break
        for tx, target_gene in reassigned:
            tx_dict[unnamed_gene].remove(tx)
            tx_dict[target_gene].append(tx)

    # if out_file is None write to stdout
    if out_file:
        out = open(out_file, "w")
    else:
        import sys

        out = sys.stdout
    # if consensus_fa is not None, write to the file
    if consensus_fa:
        out_fa = open(consensus_fa, "w")

    for tx_biotype, tx_dict in rich.progress.track(
        gene_dict_by_name.items(),
        description="Groupping genes...",
    ):
        if gene_biotype_list:
            if tx_biotype not in gene_biotype_list:
                continue
        for gene_name, tx_list in rich.progress.track(
            tx_dict.items(),
            description=f"Groupping genes ({tx_biotype})...",
        ):
            if gene_name_regex:
                if not re.search(gene_name_regex, gene_name):
                    continue

            # only cluster short genes shorter than 300bp
            if gene_length_limit:
                tx_list = [tx for tx in tx_list if tx.length <= gene_length_limit]

            names = [tx.gene_id for tx in tx_list]
            seqs = [tx.get_seq(chrom_to_fa[tx.chrom]) for tx in tx_list]
            if len(names) == 0:
                continue

            exon_spans_list = [tx.exons.values() for tx in tx_list]
            if len(tx_list) < 2:
                cluster_ids = np.arange(1, len(tx_list) + 1, dtype=int)
            else:
                msa = run_msa(names, seqs)
                aligned_array = msa_to_array(msa)
                cluster_ids = cluster_sequences(aligned_array, cluster_threshold)
            # loop the cluster ids and redo the msa for each sub-group
            for cluster_id in np.unique(cluster_ids):
                cluster_tx_list = [
                    tx for tx, cid in zip(tx_list, cluster_ids) if cid == cluster_id
                ]
                cluster_names = [
                    name for name, cid in zip(names, cluster_ids) if cid == cluster_id
                ]
                cluster_seqs = [
                    seq for seq, cid in zip(seqs, cluster_ids) if cid == cluster_id
                ]
                cluster_exon_spans_list = [
                    spans
                    for spans, cid in zip(exon_spans_list, cluster_ids)
                    if cid == cluster_id
                ]
                cluster_msa = run_msa(cluster_names, cluster_seqs, threads)
                cluster_aligned_array = msa_to_array(cluster_msa)
                cluster_consensus = consensus_sequence(cluster_aligned_array)

                # write artificial fasta file for the consensus sequences
                # header: >transcript_biotype-gene_name-cluster_id N={number of sequences in the cluster} list_of_gene_id_joined_by"|"
                # only output cluster_id less than or equal to 5, or the cluster size is greater than or equal to 3
                if consensus_fa:
                    if (
                        cluster_id <= 5
                        or len(cluster_names) >= 3
                        or gene_name in ["SNORx", "MIRx", "Ux", "RNYx"]
                        or not all(
                            n.startswith("ENS")
                            or n.startswith("LOC")
                            or n.startswith("Gm")
                            for n in cluster_names
                        )
                    ):
                        out_fa.write(
                            f">{tx_biotype}-{gene_name}-cluster{cluster_id} N={len(cluster_names)} members={'|'.join(cluster_names)}\n"
                            f"{cluster_consensus.replace('-', '').upper()}\n"
                        )

                for i, mapping in enumerate(
                    get_position_mapping_from_aligned_array(cluster_aligned_array)
                ):
                    msa_spans = []
                    for _, aligned_span in mapping:
                        msa_spans.append(aligned_span)
                    # d = map_genome_to_gap_open(cluster_exon_spans_list[i], msa_spans)
                    # d_str = ",".join([f"{k}:{v}" for k, v in d.items()])
                    # out.write(f"{gene_name}\t{cluster_id}\t{cluster_names[i]}\t{d_str}\n")
                    exon_spans_str = ",".join(
                        [
                            f"{start + 1}-{end + 1}"
                            for start, end in cluster_exon_spans_list[i]
                        ]
                    )
                    msa_spans_str = ",".join(
                        [
                            f"{start + 1}-{end + 1}"
                            for start, end in msa_spans
                            if start is not None and end is not None
                        ]
                    )
                    tx = cluster_tx_list[i]
                    tx_seq = tx.get_seq(chrom_to_fa[tx.chrom])
                    out.write(
                        f"{gene_name}\t{cluster_id}\t{len(cluster_names)}\t{cluster_consensus}\t"
                        f"{tx_biotype}\t{cluster_names[i]}\t{tx.chrom}\t{exon_spans_str}\t{tx.strand}\t{tx_seq}\t"
                        f"{msa_spans_str}\n"
                    )
    if out_file:
        out.close()
    if consensus_fa:
        out_fa.close()


if __name__ == "__main__":
    group_genes("../demo/subset.fa", "../demo/subset.tsv")



"""Tests for coralsnake.utils – Transcript, Span, helpers."""

from coralsnake.utils import (
    Span,
    Transcript,
    format_duration,
    load_annotation,
    load_faidx,
)


# ---------------------------------------------------------------------------
# Span
# ---------------------------------------------------------------------------
class TestSpan:
    def test_fields(self):
        s = Span(10, 20)
        assert s.start == 10
        assert s.end == 20

    def test_iter(self):
        assert list(Span(5, 15)) == [5, 15]

    def test_repr(self):
        assert "start=5" in repr(Span(5, 15))


# ---------------------------------------------------------------------------
# Transcript – construction and properties
# ---------------------------------------------------------------------------
class TestTranscript:
    def _make(self, strand="+"):
        return Transcript(
            gene_id="G1",
            transcript_id="T1",
            chrom="chr1",
            strand=strand,
            exons={
                1: Span(100, 200),  # 100 bp
                2: Span(300, 450),  # 150 bp
                3: Span(500, 530),  #  30 bp
            },
        )

    # --- exons_forwards ---
    def test_exons_forwards_plus(self):
        tx = self._make("+")
        fwd = tx.exons_forwards
        assert [e.start for e in fwd] == [100, 300, 500]

    def test_exons_forwards_minus(self):
        tx = self._make("-")
        fwd = tx.exons_forwards
        assert [e.start for e in fwd] == [100, 300, 500]

    # --- cum_exon_lens ---
    def test_cum_exon_lens_plus(self):
        assert self._make("+").cum_exon_lens == [100, 250, 280]

    def test_cum_exon_lens_minus(self):
        assert self._make("-").cum_exon_lens == [100, 250, 280]

    def test_cum_exon_lens_cached(self):
        tx = self._make()
        a = tx.cum_exon_lens
        b = tx.cum_exon_lens
        assert a is b

    def test_cum_exon_lens_invalidated_on_add(self):
        tx = self._make()
        _ = tx.cum_exon_lens
        tx.add_exon(4, Span(600, 650))
        assert tx.cum_exon_lens[-1] == 330

    def test_cum_exon_lens_invalidated_on_sort(self):
        tx = self._make()
        _ = tx.cum_exon_lens
        tx.sort_exons()
        assert tx._cum_exon_lens is None  # cleared by sort

    # --- length ---
    def test_length(self):
        assert self._make().length == 280

    def test_length_single_exon(self):
        tx = Transcript(exons={1: Span(0, 50)}, strand="+")
        assert tx.length == 50

    # --- to_tsv / get_genome_spans ---
    def test_to_tsv_basic(self):
        tx = self._make()
        tsv = tx.to_tsv()
        fields = tsv.split("\t")
        assert fields[0] == "G1"
        assert fields[1] == "T1"
        assert "101-200" in fields[4]

    def test_get_genome_spans(self):
        tx = self._make()
        spans = tx.get_genome_spans()
        assert "101-200" in spans

    def test_get_gene_spans(self):
        tx = self._make()
        spans = tx.get_gene_spans()
        assert spans.startswith("1-100")

    # --- repr ---
    def test_repr(self):
        tx = self._make()
        r = repr(tx)
        assert "G1" in r and "T1" in r


# ---------------------------------------------------------------------------
# load_annotation / load_faidx
# ---------------------------------------------------------------------------
class TestLoadAnnotation:
    def test_round_trip(self, tmp_path):
        tsv = tmp_path / "annot.tsv"
        tsv.write_text(
            "gene_id\ttranscript_id\tchrom\tstrand\tspans\n"
            "G1\tT1\tchr1\t+\t101-200,301-400\n"
            "G1\tT2\tchr1\t-\t501-600\n"
        )
        annot = load_annotation(str(tsv))
        assert "G1" in annot
        assert "T1" in annot["G1"]
        assert "T2" in annot["G1"]
        t1 = annot["G1"]["T1"]
        assert t1.chrom == "chr1"
        assert t1.strand == "+"
        assert len(t1.exons) == 2

    def test_no_header(self, tmp_path):
        tsv = tmp_path / "annot.tsv"
        tsv.write_text("G1\tT1\tchr1\t+\t101-200\n")
        annot = load_annotation(str(tsv), with_header=False)
        assert "T1" in annot["G1"]


class TestLoadFaidx:
    def test_basic(self, tmp_path):
        fai = tmp_path / "ref.fai"
        fai.write_text("chr1\t1000\t5\t80\t81\nchr2\t2000\t1100\t80\t81\n")
        idx = load_faidx(str(fai))
        assert idx == {"chr1": 1000, "chr2": 2000}


# ---------------------------------------------------------------------------
# format_duration
# ---------------------------------------------------------------------------
class TestFormatDuration:
    def test_seconds(self):
        assert format_duration(5.123) == "5.12s"

    def test_minutes(self):
        d = format_duration(125)
        assert "2m" in d

    def test_hours(self):
        d = format_duration(3700)
        assert "1h" in d



"""Tests for the seqops C extension module."""

from coralsnake import seqops


# ---------------------------------------------------------------------------
# reverse_complement
# ---------------------------------------------------------------------------
class TestReverseComplement:
    def test_palindrome(self):
        assert seqops.reverse_complement("ACGT") == "ACGT"

    def test_all_A(self):
        assert seqops.reverse_complement("AAAA") == "TTTT"

    def test_general(self):
        assert seqops.reverse_complement("ATCG") == "CGAT"

    def test_lowercase(self):
        assert seqops.reverse_complement("acgt") == "acgt"

    def test_N(self):
        assert seqops.reverse_complement("ANG") == "CNT"

    def test_empty(self):
        assert seqops.reverse_complement("") == ""

    def test_long(self):
        seq = "ACGT" * 250
        rc = seqops.reverse_complement(seq)
        assert len(rc) == 1000
        assert seqops.reverse_complement(rc) == seq


# ---------------------------------------------------------------------------
# base_conversion
# ---------------------------------------------------------------------------
class TestBaseConversion:
    def test_mk(self):
        # A→G, C→T (others unchanged)
        assert seqops.base_conversion("ACGT", "AC", "GT") == "GTGT"

    def test_km(self):
        # G→A, T→C (others unchanged)
        assert seqops.base_conversion("ACGT", "GT", "AC") == "ACAC"

    def test_lowercase(self):
        assert seqops.base_conversion("acgt", "AC", "GT") == "gtgt"

    def test_no_match(self):
        assert seqops.base_conversion("NNNN", "AC", "GT") == "NNNN"

    def test_empty(self):
        assert seqops.base_conversion("", "AC", "GT") == ""


# ---------------------------------------------------------------------------
# cal_md_and_tag
# ---------------------------------------------------------------------------
class TestCalMdAndTag:
    def test_perfect_match(self):
        md, yf, zf, yc, zc, ns, nc = seqops.cal_md_and_tag(
            "10M", "ACGTACGTAC", "ACGTACGTAC", True
        )
        assert md == "10"
        assert ns == 0

    def test_single_mismatch(self):
        md, *_ = seqops.cal_md_and_tag("4M", "ACGT", "ATGT", True)
        assert md == "1T2"

    def test_deletion(self):
        md, *_ = seqops.cal_md_and_tag("3M2D3M", "AAABBB", "AAAccBBB", True)
        assert "^" in md

    def test_insertion(self):
        md, yf, zf, yc, zc, ns, nc = seqops.cal_md_and_tag(
            "3M2I3M", "AAANNBBB", "AAABBB", True
        )
        assert md == "6"
        assert nc == 2  # insertion counted

    def test_soft_clip(self):
        md, yf, zf, yc, zc, ns, nc = seqops.cal_md_and_tag("2S3M", "NNACG", "ACG", True)
        assert md == "3"
        assert nc == 2

    def test_conversion_fwd(self):
        # fwd=True: b1=A, b2=G → ref=A, seq=G is a yf (expected conversion)
        md, yf, zf, yc, zc, ns, nc = seqops.cal_md_and_tag("2M", "GA", "AA", True)
        assert yf == 1
        assert zf == 1  # second base A matches, counted in zf

    def test_conversion_rev(self):
        # fwd=False: b1=T, b2=C → ref=T, seq=C is a yf (expected conversion)
        md, yf, zf, yc, zc, ns, nc = seqops.cal_md_and_tag("2M", "CT", "TT", False)
        assert yf == 1
        assert zf == 1

    def test_ns_count(self):
        # Non-conversion mismatch: ref=A, seq=C with fwd=True
        # b1=A, b2=G, b3=C, b4=T → seq[0]=C is neither b2 nor b4 → ns++
        md, yf, zf, yc, zc, ns, nc = seqops.cal_md_and_tag("1M", "C", "A", True)
        assert ns == 1


# ---------------------------------------------------------------------------
# calculate_directional_score
# ---------------------------------------------------------------------------
class TestDirectionalScore:
    def test_perfect(self):
        score, wrong, bad = seqops.calculate_directional_score(
            "10M", "ACGTACGTAC", "ACGTACGTAC", True
        )
        assert score == 10
        assert wrong == 0
        assert bad == 0

    def test_expected_conversion_o1(self):
        # Orientation 1: C→T is expected
        score, wrong, bad = seqops.calculate_directional_score("1M", "T", "C", True)
        assert score == 1
        assert bad == 0

    def test_wrong_conversion_o1(self):
        # Orientation 1: T→C is wrong
        score, wrong, bad = seqops.calculate_directional_score("1M", "C", "T", True)
        assert wrong == 1
        assert bad == 1

    def test_expected_conversion_o2(self):
        # Orientation 2: G→A is expected
        score, wrong, bad = seqops.calculate_directional_score("1M", "A", "G", False)
        assert score == 1
        assert bad == 0

    def test_wrong_conversion_o2(self):
        # Orientation 2: A→G is wrong
        score, wrong, bad = seqops.calculate_directional_score("1M", "G", "A", False)
        assert wrong == 1
        assert bad == 1

    def test_other_mismatch(self):
        # ref=A, seq=C, orientation 1 → other_mismatch
        score, wrong, bad = seqops.calculate_directional_score("1M", "C", "A", True)
        assert bad == 1
        assert wrong == 0

    def test_indel_penalty(self):
        # 3M + 1I + 3M = 6 matches - 1 indel = 5
        score, wrong, bad = seqops.calculate_directional_score(
            "3M1I3M", "AAANBBB", "AAABBB", True
        )
        assert score == 5

    def test_deletion_penalty(self):
        score, wrong, bad = seqops.calculate_directional_score(
            "3M2D3M", "AAABBB", "AAAccBBB", True
        )
        assert score == 4  # 6 matches - 2 deletion

    def test_soft_clip_no_penalty(self):
        score, wrong, bad = seqops.calculate_directional_score(
            "2S8M", "NNACGTACGT", "ACGTACGT", True
        )
        assert score == 8



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-25 14:21

import logging
import textwrap
from collections import defaultdict
from functools import lru_cache

import pysam
from . import seqops

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)


def get_logger(name: str) -> logging.Logger:
    return logging.getLogger(name)


def reverse_complement(seq: str) -> str:
    """Fast reverse complement using C implementation."""
    return seqops.reverse_complement(seq)


def mk_conversion(seq: str) -> str:
    """Convert A->G, C->T (M to K conversion)."""
    return seqops.base_conversion(seq, "AC", "GT")


def km_conversion(seq: str) -> str:
    """Convert G->A, T->C (K to M conversion)."""
    return seqops.base_conversion(seq, "GT", "AC")


def convert_file_realtime(
    input_file: str,
    output_file: str,
    base_from: str,
    base_to: str,
):
    """
    Convert FASTA file using fast C implementation (line-by-line, memory efficient).

    This function is optimized for large genome reference files. It reads and writes
    line-by-line without loading the entire file into memory, making it much faster
    and more memory-efficient than dnaio-based conversion.

    Args:
        input_file: Path to input FASTA file
        output_file: Path to output converted FASTA file
        base_from: Bases to convert from (e.g., "AC" for MK conversion)
        base_to: Bases to convert to (e.g., "GT" for MK conversion)

    Examples:
        MK conversion (A->G, C->T): convert_file_realtime(in, out, "AC", "GT")
        KM conversion (G->A, T->C): convert_file_realtime(in, out, "GT", "AC")

    Note:
        - Only works with FASTA format (not FASTQ)
        - Does not add YS:Z tags
        - Header lines (starting with '>') are written directly
        - Sequence lines are converted base-by-base
    """
    seqops.convert_fasta_file(input_file, output_file, base_from, base_to)


class Span:
    def __init__(self, start: int, end: int):
        # 0-based
        self.start = start
        self.end = end

    def __iter__(self):
        yield self.start
        yield self.end

    def __repr__(self) -> str:
        return f"(start={self.start}, end={self.end})"


class Transcript:
    def __init__(
        self,
        gene_id: str | None = None,
        transcript_id: str | None = None,
        chrom: str = "",
        strand: str = "",
        exons: dict[str | int, Span] | None = None,
        gene_name: str | None = None,
    ):
        self.gene_id = gene_id
        self.transcript_id = transcript_id
        self.chrom = chrom
        self.strand = strand
        self.exons = {} if exons is None else exons
        # calculated feature
        self._exons_forwards = None
        self._cum_exon_lens = None
        self._seq = None
        # extra feature
        self.gene_name: str | None = gene_name
        self.transcript_biotype: str | None = None
        self.start_codon: Span | None = None
        self.stop_codon: Span | None = None
        self.priority: tuple[int, int] = (10, 0)
        self.conflict: bool = False
        self._annotations: dict = {}

    def add_exon(self, exon_id: str | int, exon: Span) -> None:
        self.exons[exon_id] = exon
        self._exons_forwards = None
        self._cum_exon_lens = None
        self._seq = None

    def sort_exons(self) -> None:
        self.exons = dict(
            sorted(
                self.exons.items(), key=lambda x: x[1].start, reverse=self.strand == "-"
            ),
        )
        self._exons_forwards = None
        self._cum_exon_lens = None
        self._seq = None

    def to_tsv(
        self,
        with_codon=False,
        with_genename=False,
        with_biotype=False,
        with_txpos=False,
    ) -> str:
        # convert into 1-based
        line = []
        for key in ["gene_id", "transcript_id", "chrom", "strand"]:
            v = getattr(self, key)
            if v is not None:
                line.append(v)
            else:
                line.append("")
        line += [",".join([f"{v.start + 1}-{v.end}" for v in self.exons.values()])]
        if with_codon:
            # use the start postion of the span
            line += [
                str(self.start_codon.start + 1) if self.start_codon else "",
                str(self.stop_codon.start + 1) if self.stop_codon else "",
            ]
        if with_genename:
            line.append(self.gene_name if self.gene_name is not None else "")
        if with_biotype:
            line.append(self.transcript_biotype if self.transcript_biotype else "")
        # transcript start and end
        if with_txpos:
            _es = next(iter(self.exons.values()))
            _ee = next(reversed(self.exons.values()))
            if self.strand == "+":
                line += [str(_es.start + 1), str(_ee.end)]
            else:
                line += [str(_es.end), str(_ee.start + 1)]

        return "\t".join(line)

    def get_genome_spans(self) -> str:
        # 1-based
        return ",".join([f"{v.start + 1}-{v.end}" for v in self.exons.values()])

    def get_gene_spans(self) -> str:
        # 1-based
        gene_spans = []
        exon_start = 1
        for exon_end in self.cum_exon_lens:
            gene_spans.append(f"{exon_start}-{exon_end}")
            exon_start = exon_end + 1
        return ",".join(gene_spans)

    def get_seq(self, fasta: pysam.FastaFile, sort=True, upper=True, wrap=0):
        if self._seq is None:
            if sort:
                self.sort_exons()
            seq = ""
            for _, v in self.exons.items():
                e = fasta.fetch(self.chrom, v.start, v.end)
                if self.strand == "-":
                    e = reverse_complement(e)
                seq += e
            self._seq = seq

        seq_out = self._seq
        if upper:
            seq_out = seq_out.upper()
        if wrap > 0:
            seq_out = textwrap.fill(seq_out, wrap)
        return seq_out

    @property
    def exons_forwards(self) -> list[Span]:
        if self._exons_forwards is None:
            self._exons_forwards = sorted(self.exons.values(), key=lambda e: e.start)
        return self._exons_forwards

    @property
    def cum_exon_lens(self) -> list[int]:
        if self._cum_exon_lens is None:
            lengths = [exon.end - exon.start for exon in self.exons_forwards]
            cum_lengths = []
            total = 0
            for length in lengths:
                total += length
                cum_lengths.append(total)
            self._cum_exon_lens = cum_lengths
        return self._cum_exon_lens

    @property
    def length(self) -> int:
        return (self.cum_exon_lens or [0])[-1]

    def __repr__(self) -> str:
        res = []
        for key in [
            "gene_id",
            "transcript_id",
            "chrom",
            "strand",
            # "exons_forwards",
        ]:
            res.append(f"{key}={getattr(self, key)}")
        return f"Transcript({', '.join(res)})"


def load_annotation(
    annotation_file: str, with_header: bool = True
) -> dict[str, dict[str, Transcript]]:
    annot = defaultdict(dict)
    with open(annotation_file, "r") as f:
        if with_header:
            header = f.readline().strip("\n").split("\t")
            gene_id_idx = header.index("gene_id")
            transcript_id_idx = header.index("transcript_id")
            chrom_idx = header.index("chrom")
            strand_idx = header.index("strand")
            spans_idx = header.index("spans")
            if "gene_name" in header:
                gene_name_idx = header.index("gene_name")
            else:
                gene_name_idx = None
        else:
            (
                gene_id_idx,
                transcript_id_idx,
                chrom_idx,
                strand_idx,
                spans_idx,
                gene_name_idx,
            ) = 0, 1, 2, 3, 4, None
        for line in f:
            fields = line.strip("\n").split("\t")
            if len(fields) < 5:
                continue
            gene_id = fields[gene_id_idx]
            transcript_id = fields[transcript_id_idx]
            chrom = fields[chrom_idx]
            strand = fields[strand_idx]
            spans = fields[spans_idx]
            gene_name = fields[gene_name_idx] if gene_name_idx is not None else None
            exons = {
                idx: Span(int(start) - 1, int(end))
                for idx, span in enumerate(spans.split(","), 1)
                for start, end in [span.split("-")]
            }
            transcript = Transcript(
                gene_id=gene_id,
                transcript_id=transcript_id,
                chrom=chrom,
                strand=strand,
                exons=exons,
                gene_name=gene_name,
            )
            annot[gene_id][transcript_id] = transcript
    return annot


def load_faidx(faidx_file: str) -> dict[str, int]:
    faidx = {}
    with open(faidx_file, "r") as f:
        for line in f:
            fields = line.strip().split("\t")
            faidx[fields[0]] = int(fields[1])
    return faidx


def format_duration(sec: float) -> str:
    """
    Format duration in seconds to human-readable string.

    Args:
        sec: Duration in seconds

    Returns:
        Formatted duration string (e.g., "12.34s", "5m 23s", "2h 15m 30s")
    """
    if sec < 60:
        return f"{sec:.2f}s"
    if sec < 3600:
        minutes = int(sec // 60)
        seconds = int(round(sec % 60))
        return f"{minutes}m {seconds}s"
    hours = int(sec // 3600)
    rem = sec % 3600
    minutes = int(rem // 60)
    seconds = int(round(rem % 60))
    return f"{hours}h {minutes}m {seconds}s"



"""Tests for coralsnake.mapping – scoring helpers and integration mapping."""

from pathlib import Path

import pytest
import pysam

mapping = pytest.importorskip(
    "coralsnake.mapping", reason="mapping deps (bwamem) not available"
)

score_to_mapq = mapping.score_to_mapq
cal_md_and_tag = mapping.cal_md_and_tag
calculate_directional_score = mapping.calculate_directional_score
find_properly_paired_hits = mapping.find_properly_paired_hits
filter_hits = mapping.filter_hits


# ---------------------------------------------------------------------------
# score_to_mapq
# ---------------------------------------------------------------------------
class TestScoreToMapq:
    def test_clamped_high(self):
        assert score_to_mapq(100) == 60

    def test_clamped_low(self):
        assert score_to_mapq(-5) == 0

    def test_passthrough(self):
        assert score_to_mapq(30) == 30


# ---------------------------------------------------------------------------
# cal_md_and_tag (delegates to C)
# ---------------------------------------------------------------------------
class TestCalMdAndTag:
    def test_delegates(self):
        md, *rest = cal_md_and_tag("4M", "ACGT", "ACGT", True)
        assert md == "4"


# ---------------------------------------------------------------------------
# calculate_directional_score (delegates to C)
# ---------------------------------------------------------------------------
class TestCalculateDirectionalScore:
    def test_delegates(self):
        score, w, b = calculate_directional_score("4M", "ACGT", "ACGT", True)
        assert score == 4


# ---------------------------------------------------------------------------
# find_properly_paired_hits
# ---------------------------------------------------------------------------
class TestFindProperlyPairedHits:
    class FakeHit:
        def __init__(self, ctg, r_st, r_en, strand, read_num):
            self.ctg = ctg
            self.r_st = r_st
            self.r_en = r_en
            self.strand = strand
            self.read_num = read_num

    def test_close_pair(self):
        h1 = self.FakeHit("chr1", 100, 200, 1, 1)
        h2 = self.FakeHit("chr1", 300, 400, 1, 2)
        pairs = find_properly_paired_hits([h1, h2])
        assert len(pairs) == 1
        assert pairs[0] == (h1, h2)

    def test_too_far(self):
        h1 = self.FakeHit("chr1", 0, 100, 1, 1)
        h2 = self.FakeHit("chr1", 5000, 5100, 1, 2)
        pairs = find_properly_paired_hits([h1, h2])
        assert len(pairs) == 0

    def test_different_contigs(self):
        h1 = self.FakeHit("chr1", 100, 200, 1, 1)
        h2 = self.FakeHit("chr2", 100, 200, 1, 2)
        pairs = find_properly_paired_hits([h1, h2])
        assert len(pairs) == 0

    def test_same_read_num_not_paired(self):
        h1 = self.FakeHit("chr1", 100, 200, 1, 1)
        h2 = self.FakeHit("chr1", 300, 400, 1, 1)
        pairs = find_properly_paired_hits([h1, h2])
        assert len(pairs) == 0

    def test_symmetric_distance(self):
        h1 = self.FakeHit("chr1", 500, 600, 1, 1)
        h2 = self.FakeHit("chr1", 100, 200, 1, 2)
        pairs = find_properly_paired_hits([h1, h2])
        assert len(pairs) == 1


# ---------------------------------------------------------------------------
# filter_hits
# ---------------------------------------------------------------------------
class TestFilterHits:
    class FakeHit:
        def __init__(self, mapq, blen, mlen, read_num):
            self.mapq = mapq
            self.blen = blen
            self.mlen = mlen
            self.read_num = read_num

    def test_passes(self):
        h = self.FakeHit(mapq=10, blen=50, mlen=40, read_num=1)
        result = filter_hits(
            [h], "A" * 50, None, min_alignment_length=20, min_mapping_ratio=0.5
        )
        assert len(result) == 1

    def test_low_mapq(self):
        h = self.FakeHit(mapq=0, blen=50, mlen=40, read_num=1)
        result = filter_hits([h], "A" * 50, None)
        assert len(result) == 0

    def test_short_alignment(self):
        h = self.FakeHit(mapq=10, blen=10, mlen=10, read_num=1)
        result = filter_hits([h], "A" * 50, None, min_alignment_length=20)
        assert len(result) == 0

    def test_low_ratio(self):
        h = self.FakeHit(mapq=10, blen=50, mlen=10, read_num=1)
        result = filter_hits([h], "A" * 50, None, min_mapping_ratio=0.5)
        assert len(result) == 0


# ---------------------------------------------------------------------------
# Integration: map_file (requires bwamem at runtime)
# ---------------------------------------------------------------------------
class TestMapFileIntegration:
    @pytest.fixture(autouse=True)
    def _check_bwamem(self):
        pytest.importorskip("bwamem", reason="bwamem not installed")

    def test_pe_forward(self, tmp_path, data_dir, has_integration_data):
        out = str(tmp_path / "out.bam")
        idx = str(tmp_path / "idx")
        mapping.map_file(
            r1_file=str(data_dir / "test1.fq"),
            r2_file=str(data_dir / "test2.fq"),
            ref_files=[str(data_dir / "ref.fa")],
            output_files=out,
            forward_library=True,
            max_mismatches=0,
            threads=2,
            min_alignment_length=8,
            min_mapping_ratio=0.5,
            index_dir=idx,
            batch_size=50,
        )
        with pysam.AlignmentFile(out, "rb") as f:
            assert sum(1 for _ in f) > 0

    def test_pe_reverse(self, tmp_path, data_dir, has_integration_data):
        out = str(tmp_path / "out.bam")
        idx = str(tmp_path / "idx")
        mapping.map_file(
            r1_file=str(data_dir / "test1.fq"),
            r2_file=str(data_dir / "test2.fq"),
            ref_files=[str(data_dir / "ref.fa")],
            output_files=out,
            forward_library=False,
            max_mismatches=0,
            threads=2,
            min_alignment_length=8,
            min_mapping_ratio=0.5,
            index_dir=idx,
            batch_size=50,
        )
        with pysam.AlignmentFile(out, "rb") as f:
            assert sum(1 for _ in f) > 0

    def test_se_with_unmap(self, tmp_path, data_dir, has_integration_data):
        out = str(tmp_path / "out.bam")
        unmap = str(tmp_path / "unmap.bam")
        idx = str(tmp_path / "idx")
        mapping.map_file(
            r1_file=str(data_dir / "test1.fq"),
            r2_file=None,
            ref_files=[str(data_dir / "ref.fa")],
            output_files=out,
            unmap_file=unmap,
            forward_library=True,
            max_mismatches=0,
            threads=2,
            min_alignment_length=8,
            min_mapping_ratio=0.5,
            index_dir=idx,
            batch_size=50,
        )
        assert Path(out).exists()
        assert Path(unmap).exists()

    def test_index_only(self, tmp_path, data_dir):
        if not (data_dir / "ref.fa").exists():
            pytest.skip("ref.fa not present")
        idx = tmp_path / "idx"
        mapping.map_file(
            r1_file=None,
            r2_file=None,
            ref_files=[str(data_dir / "ref.fa")],
            output_files=str(tmp_path / "dummy.bam"),
            index_dir=str(idx),
            index_only=True,
        )
        assert (idx / "ref.orig.fa").exists()
        assert (idx / "ref.mk.amb").exists()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright © 2024 Ye Chang yech1990@gmail.com
# Distributed under terms of the GNU license.
#
# Created: 2024-06-23 18:01


import bisect
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, as_completed
from functools import lru_cache

import pysam
from rich.progress import track

from .utils import (
    Transcript,
    get_logger,
    load_annotation,
    load_faidx,
    reverse_complement,
)

LOGGER = get_logger(__name__)

COMP = str.maketrans("ACGTNacgtn", "TGCANtgcan")


@lru_cache(maxsize=10000)
def flip_flag(flag):
    if flag & 1:  # paired: toggle both read-reverse (16) and mate-reverse (32)
        return flag ^ 0x30
    else:  # unpaired: toggle only read-reverse (16)
        return flag ^ 16


@lru_cache(maxsize=10000)
def reverse_md(md: str) -> str:
    # Parse MD into components: numbers, mismatches, deletions
    parts = []
    import re

    # Match numbers, mismatches (single base), or deletions (^ followed by bases)
    for m in re.finditer(r"([0-9]+)|([A-Z])|(\^[A-Z]+)", md):
        if m.group(1) is not None:  # Number
            parts.append(int(m.group(1)))
        elif m.group(2):  # Mismatch base
            parts.append(m.group(2).translate(COMP))
        elif m.group(3):  # Deletion
            # Complement the deletion bases and reverse them
            del_bases = m.group(3)[1:]
            parts.append("^" + del_bases.translate(COMP)[::-1])

    # Reverse the parts list
    parts.reverse()

    # Merge adjacent numbers, ensuring 0s are kept between non-numbers
    merged_parts = []
    current_num = 0
    for p in parts:
        if isinstance(p, int):
            current_num += p
        else:
            merged_parts.append(str(current_num))
            current_num = 0
            merged_parts.append(p)
    merged_parts.append(str(current_num))
    return "".join(merged_parts)


@lru_cache(maxsize=100000)
def transcript_to_genome(
    transcript_pos: int, transcript: Transcript
) -> tuple[int, int]:
    """
    transcript_pos is 0-based
    """
    if transcript_pos > transcript.length or transcript_pos < 0:
        raise ValueError("Transcript position is out of range")
    # bisect_right dose not include the right bound
    # in math notation, it is [a, b)
    exon_index = bisect.bisect_right(transcript.cum_exon_lens, transcript_pos)
    if exon_index == 0:
        return transcript.exons_forwards[0].start + transcript_pos, 0
    offset = transcript_pos - transcript.cum_exon_lens[exon_index - 1]
    return transcript.exons_forwards[exon_index].start + offset, exon_index


def remap_to_genome(
    align: pysam.AlignedSegment,
    header: pysam.AlignmentHeader,
    transcript: Transcript,
    next_transcript: Transcript,
) -> pysam.AlignedSegment:
    new_align = pysam.AlignedSegment(header=header)
    new_align.query_name = align.query_name
    new_align.flag = align.flag

    if align.cigartuples is None:
        raise ValueError("CIGAR string is required")

    for tag in align.get_tags():
        new_align.set_tag(*tag)
    if transcript.strand == "-":
        new_align.flag = flip_flag(align.flag)
        new_align.query_sequence = reverse_complement(align.query_sequence)
        new_align.query_qualities = (
            align.query_qualities[::-1] if align.query_qualities else None
        )
        cigartuples = align.cigartuples[::-1]
        if new_align.has_tag("MD"):
            new_align.set_tag("MD", reverse_md(new_align.get_tag("MD")))
        transcript_pos = transcript.length - align.reference_end
    else:
        new_align.query_sequence = align.query_sequence
        new_align.query_qualities = align.query_qualities
        cigartuples = align.cigartuples
        transcript_pos = align.reference_start

    genome_pos, exon_index = transcript_to_genome(transcript_pos, transcript)
    new_align.reference_name = transcript.chrom
    if next_transcript is not None:
        new_align.next_reference_name = next_transcript.chrom
    new_align.reference_start = genome_pos

    new_cigar = []
    for op, length in cigartuples:
        if op in (0, 2):  # Match, deletion
            while length > 0:
                current_exon = transcript.exons_forwards[exon_index]
                exon_remaining = current_exon.end - genome_pos
                if length <= exon_remaining:
                    new_cigar.append((op, length))
                    genome_pos += length
                    length = 0
                else:
                    if exon_remaining > 0:
                        new_cigar.append((op, exon_remaining))
                        genome_pos += exon_remaining
                        length -= exon_remaining
                    if exon_index + 1 < len(transcript.exons_forwards):
                        next_exon = transcript.exons_forwards[exon_index + 1]
                        intron_length = next_exon.start - current_exon.end
                        new_cigar.append((3, intron_length))
                        genome_pos = next_exon.start
                        exon_index += 1
                    else:
                        break
        else:  # Insertion or soft-clip
            new_cigar.append((op, length))

    new_align.cigartuples = new_cigar
    new_align.mapping_quality = align.mapping_quality

    return new_align


def parse_alignment(
    align: pysam.AlignedSegment,
    annot: dict[str, Transcript],
    genome_header: pysam.AlignmentHeader,
) -> pysam.AlignedSegment:
    if (
        align.is_unmapped
        or align.reference_name is None
        or align.reference_name not in annot
    ):
        align.reference_id = -1
        align.reference_start = -1
        align.next_reference_id = -1
        align.next_reference_start = -1
        return align

    transcript = annot[align.reference_name]
    if align.next_reference_name in annot:
        next_transcript = annot[align.next_reference_name]
    else:
        next_transcript = None
    new_align = remap_to_genome(align, genome_header, transcript, next_transcript)
    return new_align


# Worker global state for multiprocessing
_WORKER_ANNOT = None
_WORKER_GENOME_HEADER = None
_WORKER_TRANSCRIPT_HEADER = None


def _init_worker(annotation_file, genome_header_dict, transcript_header_dict):
    """Initialize worker with shared annotation and headers."""
    global _WORKER_ANNOT, _WORKER_GENOME_HEADER, _WORKER_TRANSCRIPT_HEADER
    raw_annot = load_annotation(annotation_file)
    # Build flat mapping for transcript lookups
    _WORKER_ANNOT = {}
    for g_id in raw_annot:
        for t_id, tx in raw_annot[g_id].items():
            _WORKER_ANNOT[t_id] = tx

    _WORKER_GENOME_HEADER = pysam.AlignmentHeader.from_dict(genome_header_dict)
    _WORKER_TRANSCRIPT_HEADER = pysam.AlignmentHeader.from_dict(transcript_header_dict)


def _process_chunk(align_strings):
    """Process a chunk of alignments (serialized as strings)."""
    results = []
    for s in align_strings:
        # Deserialize from string using transcript header
        align = pysam.AlignedSegment.fromstring(s, _WORKER_TRANSCRIPT_HEADER)
        # Remap to genome
        new_align = parse_alignment(align, _WORKER_ANNOT, _WORKER_GENOME_HEADER)
        # Serialize result to string for return
        results.append(new_align.to_string())
    return results


def convert_bam(
    input_bam: str,
    output_bam: str,
    annotation_file: str,
    faidx_file: str,
    threads: int = 8,
    sort: bool = False,
):
    LOGGER.info("Loading fasta index...")
    faidx = load_faidx(faidx_file)

    with pysam.AlignmentFile(input_bam, "rb") as in_bam:
        transcript_header_dict = in_bam.header.to_dict()
        new_header = in_bam.header.to_dict()
        if "HD" not in new_header:
            new_header["HD"] = {"VN": "1.4"}
        # mark as unsorted
        new_header["HD"]["SO"] = "unsorted"
        new_header["SQ"] = [
            {"SN": chrom, "LN": length} for chrom, length in faidx.items()
        ]
        genome_header_dict = new_header
        genome_header = pysam.AlignmentHeader.from_dict(genome_header_dict)

        with pysam.AlignmentFile(
            output_bam,
            "wb" if output_bam.endswith(".bam") else "w",
            header=genome_header,
        ) as out_bam:
            if threads <= 1:
                LOGGER.info("Loading annotation for single-threaded processing...")
                raw_annot = load_annotation(annotation_file)
                annot = {}
                for g_id in raw_annot:
                    for t_id, tx in raw_annot[g_id].items():
                        annot[t_id] = tx
                for align in track(in_bam, description="Processing..."):
                    new_align = parse_alignment(align, annot, genome_header)
                    out_bam.write(new_align)
            else:
                LOGGER.info(f"Using {threads} workers for parallel processing...")
                chunk_size = 5000
                # Keep a limited number of futures in flight to save memory
                max_queue = threads * 2
                futures = []
                chunk = []

                with ProcessPoolExecutor(
                    max_workers=threads,
                    mp_context=mp.get_context("spawn"),
                    initializer=_init_worker,
                    initargs=(
                        annotation_file,
                        genome_header_dict,
                        transcript_header_dict,
                    ),
                ) as executor:
                    for align in track(in_bam, description="Processing..."):
                        chunk.append(align.to_string())
                        if len(chunk) >= chunk_size:
                            futures.append(executor.submit(_process_chunk, list(chunk)))
                            chunk = []

                            # Drain results if queue is full
                            if len(futures) >= max_queue:
                                for res_str in futures.pop(0).result():
                                    res_align = pysam.AlignedSegment.fromstring(
                                        res_str, genome_header
                                    )
                                    out_bam.write(res_align)

                    # Submit final chunk
                    if chunk:
                        futures.append(executor.submit(_process_chunk, chunk))

                    # Drain remaining futures in order to maintain read sequence
                    for fut in futures:
                        for res_str in fut.result():
                            res_align = pysam.AlignedSegment.fromstring(
                                res_str, genome_header
                            )
                            out_bam.write(res_align)

    if sort:
        # skip sorting if output is not BAM file
        if not output_bam.endswith(".bam"):
            LOGGER.info("Output is not BAM file, skip sorting and indexing")
            return
        LOGGER.info("Sorting output BAM file...")
        import shutil
        import uuid

        internal_sorted_bam = f".{str(uuid.uuid4())[:8]}_{output_bam}"
        pysam.sort("-@", str(threads), "-o", internal_sorted_bam, output_bam)
        shutil.move(internal_sorted_bam, output_bam)
        pysam.index("-@", str(threads), output_bam)


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Remap transcriptome BAM to genome BAM"
    )
    parser.add_argument("-i", "--input", required=True, help="Input BAM file")
    parser.add_argument("-o", "--output", required=True, help="Output BAM file")
    parser.add_argument("-a", "--annotation", required=True, help="Annotation file")
    parser.add_argument("-f", "--faidx", required=True, help="Fasta index file")
    # sort bam file
    parser.add_argument(
        "--sort", action="store_true", help="Sort output BAM file by coordinates"
    )
    parser.add_argument(
        "-t", "--threads", type=int, default=8, help="Number of threads"
    )
    args = parser.parse_args()

    convert_bam(
        args.input, args.output, args.annotation, args.faidx, args.threads, args.sort
    )



"""Tests for coralsnake.annot – site annotation with transcript positions."""

from pathlib import Path

import pytest

run_annot = pytest.importorskip(
    "coralsnake.annot", reason="annot deps not available", exc_type=ImportError
).run_annot


def _write(p: Path, content: str) -> str:
    p.write_text(content)
    return str(p)


# ---------------------------------------------------------------------------
# run_annot basics
# ---------------------------------------------------------------------------
class TestRunAnnot:
    def _annot(self, tmp_path):
        return _write(
            tmp_path / "annot.tsv",
            "chrom\tstrand\tspans\tgene_id\ttranscript_id\n"
            "chr1\t+\t10-20,30-40\tGENE1\tTX1\n"
            "chr1\t-\t50-60\tGENE2\tTX2\n",
        )

    def test_plus_strand_first_base(self, tmp_path):
        annot = self._annot(tmp_path)
        inp = _write(tmp_path / "sites.tsv", "chr1\t10\t+\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot, keep_na=True)
        line = Path(out).read_text().strip().split("\n")[0]
        assert line == "chr1\t10\t+\tGENE1\tTX1\t0"

    def test_plus_strand_second_exon(self, tmp_path):
        annot = self._annot(tmp_path)
        inp = _write(tmp_path / "sites.tsv", "chr1\t35\t+\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot, keep_na=True)
        line = Path(out).read_text().strip().split("\n")[0]
        assert line == "chr1\t35\t+\tGENE1\tTX1\t16"

    def test_no_hit_gives_na(self, tmp_path):
        annot = self._annot(tmp_path)
        inp = _write(tmp_path / "sites.tsv", "chr1\t5\t+\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot, keep_na=True)
        line = Path(out).read_text().strip().split("\n")[0]
        assert line == "chr1\t5\t+\t.\t.\t."

    def test_keep_na_false(self, tmp_path):
        annot = self._annot(tmp_path)
        inp = _write(tmp_path / "sites.tsv", "chr1\t5\t+\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot, keep_na=False)
        assert Path(out).read_text().strip() == ""

    def test_minus_strand_position(self, tmp_path):
        """Verify the off-by-one fix: exon_end - 1 - position + exon_shift."""
        annot = self._annot(tmp_path)
        inp = _write(tmp_path / "sites.tsv", "chr1\t60\t-\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot, keep_na=True)
        line = Path(out).read_text().strip().split("\n")[0]
        assert line == "chr1\t60\t-\tGENE2\tTX2\t0"

    def test_minus_strand_last_base(self, tmp_path):
        annot = self._annot(tmp_path)
        inp = _write(tmp_path / "sites.tsv", "chr1\t51\t-\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot, keep_na=True)
        line = Path(out).read_text().strip().split("\n")[0]
        assert line == "chr1\t51\t-\tGENE2\tTX2\t9"

    def test_collapse_annot(self, tmp_path):
        annot_path = _write(
            tmp_path / "annot.tsv",
            "chrom\tstrand\tspans\tgene_id\ttranscript_id\n"
            "chr1\t+\t10-20\tG1\tT1\n"
            "chr1\t+\t10-20\tG2\tT2\n",
        )
        inp = _write(tmp_path / "sites.tsv", "chr1\t15\t+\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot_path, keep_na=True, collapse_annot=True)
        line = Path(out).read_text().strip().split("\n")[0]
        assert "G1,G2" in line or "G2,G1" in line

    def test_add_count(self, tmp_path):
        annot_path = _write(
            tmp_path / "annot.tsv",
            "chrom\tstrand\tspans\tgene_id\ttranscript_id\n"
            "chr1\t+\t10-20\tG1\tT1\n"
            "chr1\t+\t10-20\tG2\tT2\n",
        )
        inp = _write(tmp_path / "sites.tsv", "chr1\t15\t+\n")
        out = str(tmp_path / "out.tsv")
        run_annot(inp, out, annot_path, keep_na=True, add_count=True)
        lines = Path(out).read_text().strip().split("\n")
        assert all(line.endswith("\t2") for line in lines)



"""Tests for coralsnake.tbam2gbam – reverse_md, flip_flag, transcript_to_genome, remap_to_genome."""

import pysam
import pytest

from coralsnake.tbam2gbam import (
    flip_flag,
    remap_to_genome,
    reverse_md,
    transcript_to_genome,
)
from coralsnake.utils import Span, Transcript


# ---------------------------------------------------------------------------
# reverse_md
# ---------------------------------------------------------------------------
class TestReverseMd:
    def test_all_match(self):
        assert reverse_md("10") == "10"

    def test_zero(self):
        assert reverse_md("0") == "0"

    def test_single_mismatch(self):
        assert reverse_md("2A3") == "3T2"

    def test_adjacent_mismatches(self):
        assert reverse_md("0A0C5") == "5G0T0"

    def test_deletion(self):
        # ^ACG → comp(ACG)=TGC → rev(TGC)=CGT → ^CGT
        assert reverse_md("2^ACG3") == "3^CGT2"

    def test_mismatch_then_deletion(self):
        assert reverse_md("0A0^GT3") == "3^AC0T0"

    def test_pure_mismatch(self):
        assert reverse_md("0G0") == "0C0"

    def test_complex(self):
        assert reverse_md("3A2^TT0C4") == "4G0^AA2T3"

    def test_symmetric_deletion(self):
        # ^ACGT: comp=TGCA, rev=ACGT → same
        assert reverse_md("0^ACGT0") == "0^ACGT0"

    def test_cache_identity(self):
        a = reverse_md("5A5")
        b = reverse_md("5A5")
        assert a is b


# ---------------------------------------------------------------------------
# flip_flag
# ---------------------------------------------------------------------------
class TestFlipFlag:
    def test_unpaired_fwd(self):
        assert flip_flag(0) == 16

    def test_unpaired_rev(self):
        assert flip_flag(16) == 0

    def test_paired_99(self):
        assert flip_flag(99) == 83

    def test_paired_147(self):
        assert flip_flag(147) == 163

    def test_paired_both_fwd(self):
        assert flip_flag(67) == 67 ^ 0x30

    def test_paired_both_rev(self):
        f = 1 | 0x10 | 0x20
        assert flip_flag(f) == f ^ 0x30

    def test_preserves_other_bits_paired(self):
        assert flip_flag(67) & ~0x30 == 67 & ~0x30

    def test_preserves_other_bits_unpaired(self):
        assert flip_flag(272) == 256

    def test_double_flip_identity(self):
        for f in [0, 16, 99, 147, 67, 131, 4, 256]:
            assert flip_flag(flip_flag(f)) == f


# ---------------------------------------------------------------------------
# transcript_to_genome
# ---------------------------------------------------------------------------
class TestTranscriptToGenome:
    def _tx(self, strand="+"):
        return Transcript(
            gene_id="G",
            transcript_id="T",
            chrom="chr1",
            strand=strand,
            exons={1: Span(1000, 1100), 2: Span(2000, 2050), 3: Span(3000, 3030)},
        )

    def test_first_exon_start(self):
        pos, idx = transcript_to_genome(0, self._tx())
        assert (pos, idx) == (1000, 0)

    def test_first_exon_end(self):
        pos, idx = transcript_to_genome(99, self._tx())
        assert (pos, idx) == (1099, 0)

    def test_second_exon_boundary(self):
        pos, idx = transcript_to_genome(100, self._tx())
        assert (pos, idx) == (2000, 1)

    def test_second_exon_mid(self):
        pos, idx = transcript_to_genome(125, self._tx())
        assert (pos, idx) == (2025, 1)

    def test_third_exon(self):
        pos, idx = transcript_to_genome(150, self._tx())
        assert (pos, idx) == (3000, 2)

    def test_minus_strand(self):
        tx = self._tx("-")
        pos, idx = transcript_to_genome(0, tx)
        assert pos == 1000 and idx == 0

    def test_minus_strand_second_exon(self):
        tx = self._tx("-")
        pos, idx = transcript_to_genome(100, tx)
        assert pos == 2000 and idx == 1

    def test_out_of_range(self):
        with pytest.raises(ValueError):
            transcript_to_genome(999, self._tx())
        with pytest.raises(ValueError):
            transcript_to_genome(-1, self._tx())

    def test_single_exon(self):
        tx = Transcript(
            gene_id="G",
            transcript_id="T",
            chrom="chr1",
            strand="+",
            exons={1: Span(500, 600)},
        )
        pos, idx = transcript_to_genome(50, tx)
        assert (pos, idx) == (550, 0)


# ---------------------------------------------------------------------------
# remap_to_genome (liftover)
# ---------------------------------------------------------------------------
class TestRemapToGenome:
    def _genome_header(self):
        return pysam.AlignmentHeader.from_dict(
            {
                "HD": {"VN": "1.6"},
                "SQ": [{"SN": "chr1", "LN": 10000}],
            }
        )

    def _tx_header(self):
        return pysam.AlignmentHeader.from_dict(
            {
                "HD": {"VN": "1.6"},
                "SQ": [{"SN": "TX1", "LN": 300}, {"SN": "TX2", "LN": 300}],
            }
        )

    def _plus_tx(self):
        return Transcript(
            gene_id="G1",
            transcript_id="TX1",
            chrom="chr1",
            strand="+",
            exons={1: Span(1000, 1050), 2: Span(2000, 2050)},
        )

    def _minus_tx(self):
        return Transcript(
            gene_id="G2",
            transcript_id="TX2",
            chrom="chr1",
            strand="-",
            exons={1: Span(1000, 1050), 2: Span(2000, 2050)},
        )

    def _make_align(self, ref_name, start, cigar, flag=0, seq="ACGTACGTAC", md="10"):
        a = pysam.AlignedSegment(header=self._tx_header())
        a.query_name = "read"
        a.flag = flag
        a.reference_name = ref_name
        a.reference_start = start
        a.query_sequence = seq
        a.query_qualities = pysam.qualitystring_to_array("I" * len(seq))
        a.cigarstring = cigar
        a.mapping_quality = 60
        a.set_tag("MD", md)
        return a

    def test_plus_simple(self):
        a = self._make_align("TX1", 10, "10M")
        r = remap_to_genome(a, self._genome_header(), self._plus_tx(), None)
        assert r.reference_start == 1010
        assert r.flag == 0

    def test_plus_intron(self):
        a = self._make_align("TX1", 45, "10M")
        r = remap_to_genome(a, self._genome_header(), self._plus_tx(), None)
        assert r.reference_start == 1045
        ops = r.cigartuples
        assert ops[0] == (0, 5)
        assert ops[1] == (3, 950)
        assert ops[2] == (0, 5)

    def test_minus_flips_flag_unpaired(self):
        a = self._make_align("TX2", 10, "10M", flag=0)
        r = remap_to_genome(a, self._genome_header(), self._minus_tx(), None)
        assert r.flag == 16

    def test_minus_flips_flag_paired(self):
        a = self._make_align("TX2", 10, "10M", flag=99)
        r = remap_to_genome(a, self._genome_header(), self._minus_tx(), None)
        assert r.flag == 83

    def test_minus_reverses_seq(self):
        a = self._make_align("TX2", 0, "6M", seq="AAACCC", md="6")
        a.query_qualities = pysam.qualitystring_to_array("ABCDEF")
        r = remap_to_genome(a, self._genome_header(), self._minus_tx(), None)
        assert r.query_sequence == "GGGTTT"
        assert list(r.query_qualities) == list(pysam.qualitystring_to_array("FEDCBA"))

    def test_minus_reverses_cigar(self):
        a = self._make_align("TX2", 0, "2S8M", seq="AAACGTACGT", md="8")
        r = remap_to_genome(a, self._genome_header(), self._minus_tx(), None)
        assert r.cigartuples[0][0] == 0  # M first
        assert r.cigartuples[-1][0] == 4  # S last

    def test_minus_intron_spanning(self):
        """A read near the exon junction on minus strand should produce M-N-M on the genome."""
        a = self._make_align("TX2", 45, "10M", seq="ACGTACGTAC", md="10")
        r = remap_to_genome(a, self._genome_header(), self._minus_tx(), None)
        ops = r.cigartuples
        assert ops[0] == (0, 5)  # 5M in first genomic exon
        assert ops[1] == (3, 950)  # 950N intron
        assert ops[2] == (0, 5)  # 5M in second genomic exon


# ---------------------------------------------------------------------------
# Integration: convert_bam
# ---------------------------------------------------------------------------
class TestConvertBam:
    def test_liftover_basic(self, tmp_path, data_dir, has_liftover_data):
        from pathlib import Path
        from coralsnake.tbam2gbam import convert_bam

        output = str(tmp_path / "output.bam")
        convert_bam(
            input_bam=str(data_dir / "liftover_input.bam"),
            output_bam=output,
            annotation_file=str(data_dir / "liftover_annot.tsv"),
            faidx_file=str(data_dir / "liftover_faidx.fai"),
            threads=1,
            sort=False,
        )
        assert Path(output).exists()
        with pysam.AlignmentFile(output, "rb") as f:
            assert f.header is not None
